# Generated from \CSharpParser.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CSharpParser import CSharpParser
else:
    from CSharpParser import CSharpParser

# This class defines a complete listener for a parse tree produced by CSharpParser.


class CSharpParserListener(ParseTreeListener):

    # Enter a parse tree produced by CSharpParser#compilation_unit.
    def enterCompilation_unit(self, ctx: CSharpParser.Compilation_unitContext):
        pass

    # Exit a parse tree produced by CSharpParser#compilation_unit.
    def exitCompilation_unit(self, ctx: CSharpParser.Compilation_unitContext):
        pass

    # Enter a parse tree produced by CSharpParser#namespace_or_type_name.
    def enterNamespace_or_type_name(self, ctx: CSharpParser.Namespace_or_type_nameContext):
        pass

    # Exit a parse tree produced by CSharpParser#namespace_or_type_name.
    def exitNamespace_or_type_name(self, ctx: CSharpParser.Namespace_or_type_nameContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_.
    def enterType_(self, ctx: CSharpParser.Type_Context):
        pass

    # Exit a parse tree produced by CSharpParser#type_.
    def exitType_(self, ctx: CSharpParser.Type_Context):
        pass

    # Enter a parse tree produced by CSharpParser#base_type.
    def enterBase_type(self, ctx: CSharpParser.Base_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#base_type.
    def exitBase_type(self, ctx: CSharpParser.Base_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#tuple_type.
    def enterTuple_type(self, ctx: CSharpParser.Tuple_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#tuple_type.
    def exitTuple_type(self, ctx: CSharpParser.Tuple_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#tuple_element.
    def enterTuple_element(self, ctx: CSharpParser.Tuple_elementContext):
        pass

    # Exit a parse tree produced by CSharpParser#tuple_element.
    def exitTuple_element(self, ctx: CSharpParser.Tuple_elementContext):
        pass

    # Enter a parse tree produced by CSharpParser#simple_type.
    def enterSimple_type(self, ctx: CSharpParser.Simple_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#simple_type.
    def exitSimple_type(self, ctx: CSharpParser.Simple_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#numeric_type.
    def enterNumeric_type(self, ctx: CSharpParser.Numeric_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#numeric_type.
    def exitNumeric_type(self, ctx: CSharpParser.Numeric_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#integral_type.
    def enterIntegral_type(self, ctx: CSharpParser.Integral_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#integral_type.
    def exitIntegral_type(self, ctx: CSharpParser.Integral_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#floating_point_type.
    def enterFloating_point_type(self, ctx: CSharpParser.Floating_point_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#floating_point_type.
    def exitFloating_point_type(self, ctx: CSharpParser.Floating_point_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_type.
    def enterClass_type(self, ctx: CSharpParser.Class_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_type.
    def exitClass_type(self, ctx: CSharpParser.Class_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_argument_list.
    def enterType_argument_list(self, ctx: CSharpParser.Type_argument_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_argument_list.
    def exitType_argument_list(self, ctx: CSharpParser.Type_argument_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#argument_list.
    def enterArgument_list(self, ctx: CSharpParser.Argument_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#argument_list.
    def exitArgument_list(self, ctx: CSharpParser.Argument_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#argument.
    def enterArgument(self, ctx: CSharpParser.ArgumentContext):
        pass

    # Exit a parse tree produced by CSharpParser#argument.
    def exitArgument(self, ctx: CSharpParser.ArgumentContext):
        pass

    # Enter a parse tree produced by CSharpParser#expression.
    def enterExpression(self, ctx: CSharpParser.ExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#expression.
    def exitExpression(self, ctx: CSharpParser.ExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#non_assignment_expression.
    def enterNon_assignment_expression(self, ctx: CSharpParser.Non_assignment_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#non_assignment_expression.
    def exitNon_assignment_expression(self, ctx: CSharpParser.Non_assignment_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#assignment.
    def enterAssignment(self, ctx: CSharpParser.AssignmentContext):
        pass

    # Exit a parse tree produced by CSharpParser#assignment.
    def exitAssignment(self, ctx: CSharpParser.AssignmentContext):
        pass

    # Enter a parse tree produced by CSharpParser#assignment_operator.
    def enterAssignment_operator(self, ctx: CSharpParser.Assignment_operatorContext):
        pass

    # Exit a parse tree produced by CSharpParser#assignment_operator.
    def exitAssignment_operator(self, ctx: CSharpParser.Assignment_operatorContext):
        pass

    # Enter a parse tree produced by CSharpParser#conditional_expression.
    def enterConditional_expression(self, ctx: CSharpParser.Conditional_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#conditional_expression.
    def exitConditional_expression(self, ctx: CSharpParser.Conditional_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#null_coalescing_expression.
    def enterNull_coalescing_expression(self, ctx: CSharpParser.Null_coalescing_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#null_coalescing_expression.
    def exitNull_coalescing_expression(self, ctx: CSharpParser.Null_coalescing_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#conditional_or_expression.
    def enterConditional_or_expression(self, ctx: CSharpParser.Conditional_or_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#conditional_or_expression.
    def exitConditional_or_expression(self, ctx: CSharpParser.Conditional_or_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#conditional_and_expression.
    def enterConditional_and_expression(self, ctx: CSharpParser.Conditional_and_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#conditional_and_expression.
    def exitConditional_and_expression(self, ctx: CSharpParser.Conditional_and_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#inclusive_or_expression.
    def enterInclusive_or_expression(self, ctx: CSharpParser.Inclusive_or_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#inclusive_or_expression.
    def exitInclusive_or_expression(self, ctx: CSharpParser.Inclusive_or_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#exclusive_or_expression.
    def enterExclusive_or_expression(self, ctx: CSharpParser.Exclusive_or_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#exclusive_or_expression.
    def exitExclusive_or_expression(self, ctx: CSharpParser.Exclusive_or_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#and_expression.
    def enterAnd_expression(self, ctx: CSharpParser.And_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#and_expression.
    def exitAnd_expression(self, ctx: CSharpParser.And_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#equality_expression.
    def enterEquality_expression(self, ctx: CSharpParser.Equality_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#equality_expression.
    def exitEquality_expression(self, ctx: CSharpParser.Equality_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#relational_expression.
    def enterRelational_expression(self, ctx: CSharpParser.Relational_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#relational_expression.
    def exitRelational_expression(self, ctx: CSharpParser.Relational_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#shift_expression.
    def enterShift_expression(self, ctx: CSharpParser.Shift_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#shift_expression.
    def exitShift_expression(self, ctx: CSharpParser.Shift_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#additive_expression.
    def enterAdditive_expression(self, ctx: CSharpParser.Additive_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#additive_expression.
    def exitAdditive_expression(self, ctx: CSharpParser.Additive_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#multiplicative_expression.
    def enterMultiplicative_expression(self, ctx: CSharpParser.Multiplicative_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#multiplicative_expression.
    def exitMultiplicative_expression(self, ctx: CSharpParser.Multiplicative_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#switch_expression.
    def enterSwitch_expression(self, ctx: CSharpParser.Switch_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#switch_expression.
    def exitSwitch_expression(self, ctx: CSharpParser.Switch_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#switch_expression_arms.
    def enterSwitch_expression_arms(self, ctx: CSharpParser.Switch_expression_armsContext):
        pass

    # Exit a parse tree produced by CSharpParser#switch_expression_arms.
    def exitSwitch_expression_arms(self, ctx: CSharpParser.Switch_expression_armsContext):
        pass

    # Enter a parse tree produced by CSharpParser#switch_expression_arm.
    def enterSwitch_expression_arm(self, ctx: CSharpParser.Switch_expression_armContext):
        pass

    # Exit a parse tree produced by CSharpParser#switch_expression_arm.
    def exitSwitch_expression_arm(self, ctx: CSharpParser.Switch_expression_armContext):
        pass

    # Enter a parse tree produced by CSharpParser#range_expression.
    def enterRange_expression(self, ctx: CSharpParser.Range_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#range_expression.
    def exitRange_expression(self, ctx: CSharpParser.Range_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#unary_expression.
    def enterUnary_expression(self, ctx: CSharpParser.Unary_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#unary_expression.
    def exitUnary_expression(self, ctx: CSharpParser.Unary_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#primary_expression.
    def enterPrimary_expression(self, ctx: CSharpParser.Primary_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#primary_expression.
    def exitPrimary_expression(self, ctx: CSharpParser.Primary_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#literalExpression.
    def enterLiteralExpression(self, ctx: CSharpParser.LiteralExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#literalExpression.
    def exitLiteralExpression(self, ctx: CSharpParser.LiteralExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#simpleNameExpression.
    def enterSimpleNameExpression(self, ctx: CSharpParser.SimpleNameExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#simpleNameExpression.
    def exitSimpleNameExpression(self, ctx: CSharpParser.SimpleNameExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#parenthesisExpressions.
    def enterParenthesisExpressions(self, ctx: CSharpParser.ParenthesisExpressionsContext):
        pass

    # Exit a parse tree produced by CSharpParser#parenthesisExpressions.
    def exitParenthesisExpressions(self, ctx: CSharpParser.ParenthesisExpressionsContext):
        pass

    # Enter a parse tree produced by CSharpParser#memberAccessExpression.
    def enterMemberAccessExpression(self, ctx: CSharpParser.MemberAccessExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#memberAccessExpression.
    def exitMemberAccessExpression(self, ctx: CSharpParser.MemberAccessExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#literalAccessExpression.
    def enterLiteralAccessExpression(self, ctx: CSharpParser.LiteralAccessExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#literalAccessExpression.
    def exitLiteralAccessExpression(self, ctx: CSharpParser.LiteralAccessExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#thisReferenceExpression.
    def enterThisReferenceExpression(self, ctx: CSharpParser.ThisReferenceExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#thisReferenceExpression.
    def exitThisReferenceExpression(self, ctx: CSharpParser.ThisReferenceExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#baseAccessExpression.
    def enterBaseAccessExpression(self, ctx: CSharpParser.BaseAccessExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#baseAccessExpression.
    def exitBaseAccessExpression(self, ctx: CSharpParser.BaseAccessExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#objectCreationExpression.
    def enterObjectCreationExpression(self, ctx: CSharpParser.ObjectCreationExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#objectCreationExpression.
    def exitObjectCreationExpression(self, ctx: CSharpParser.ObjectCreationExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#tupleExpression.
    def enterTupleExpression(self, ctx: CSharpParser.TupleExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#tupleExpression.
    def exitTupleExpression(self, ctx: CSharpParser.TupleExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#typeofExpression.
    def enterTypeofExpression(self, ctx: CSharpParser.TypeofExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#typeofExpression.
    def exitTypeofExpression(self, ctx: CSharpParser.TypeofExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#checkedExpression.
    def enterCheckedExpression(self, ctx: CSharpParser.CheckedExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#checkedExpression.
    def exitCheckedExpression(self, ctx: CSharpParser.CheckedExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#uncheckedExpression.
    def enterUncheckedExpression(self, ctx: CSharpParser.UncheckedExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#uncheckedExpression.
    def exitUncheckedExpression(self, ctx: CSharpParser.UncheckedExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#defaultValueExpression.
    def enterDefaultValueExpression(self, ctx: CSharpParser.DefaultValueExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#defaultValueExpression.
    def exitDefaultValueExpression(self, ctx: CSharpParser.DefaultValueExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#anonymousMethodExpression.
    def enterAnonymousMethodExpression(self, ctx: CSharpParser.AnonymousMethodExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#anonymousMethodExpression.
    def exitAnonymousMethodExpression(self, ctx: CSharpParser.AnonymousMethodExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#sizeofExpression.
    def enterSizeofExpression(self, ctx: CSharpParser.SizeofExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#sizeofExpression.
    def exitSizeofExpression(self, ctx: CSharpParser.SizeofExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#nameofExpression.
    def enterNameofExpression(self, ctx: CSharpParser.NameofExpressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#nameofExpression.
    def exitNameofExpression(self, ctx: CSharpParser.NameofExpressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#throwable_expression.
    def enterThrowable_expression(self, ctx: CSharpParser.Throwable_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#throwable_expression.
    def exitThrowable_expression(self, ctx: CSharpParser.Throwable_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#throw_expression.
    def enterThrow_expression(self, ctx: CSharpParser.Throw_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#throw_expression.
    def exitThrow_expression(self, ctx: CSharpParser.Throw_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_access.
    def enterMember_access(self, ctx: CSharpParser.Member_accessContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_access.
    def exitMember_access(self, ctx: CSharpParser.Member_accessContext):
        pass

    # Enter a parse tree produced by CSharpParser#bracket_expression.
    def enterBracket_expression(self, ctx: CSharpParser.Bracket_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#bracket_expression.
    def exitBracket_expression(self, ctx: CSharpParser.Bracket_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#indexer_argument.
    def enterIndexer_argument(self, ctx: CSharpParser.Indexer_argumentContext):
        pass

    # Exit a parse tree produced by CSharpParser#indexer_argument.
    def exitIndexer_argument(self, ctx: CSharpParser.Indexer_argumentContext):
        pass

    # Enter a parse tree produced by CSharpParser#predefined_type.
    def enterPredefined_type(self, ctx: CSharpParser.Predefined_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#predefined_type.
    def exitPredefined_type(self, ctx: CSharpParser.Predefined_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#expression_list.
    def enterExpression_list(self, ctx: CSharpParser.Expression_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#expression_list.
    def exitExpression_list(self, ctx: CSharpParser.Expression_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#object_or_collection_initializer.
    def enterObject_or_collection_initializer(self, ctx: CSharpParser.Object_or_collection_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#object_or_collection_initializer.
    def exitObject_or_collection_initializer(self, ctx: CSharpParser.Object_or_collection_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#object_initializer.
    def enterObject_initializer(self, ctx: CSharpParser.Object_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#object_initializer.
    def exitObject_initializer(self, ctx: CSharpParser.Object_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_initializer_list.
    def enterMember_initializer_list(self, ctx: CSharpParser.Member_initializer_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_initializer_list.
    def exitMember_initializer_list(self, ctx: CSharpParser.Member_initializer_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_initializer.
    def enterMember_initializer(self, ctx: CSharpParser.Member_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_initializer.
    def exitMember_initializer(self, ctx: CSharpParser.Member_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#initializer_value.
    def enterInitializer_value(self, ctx: CSharpParser.Initializer_valueContext):
        pass

    # Exit a parse tree produced by CSharpParser#initializer_value.
    def exitInitializer_value(self, ctx: CSharpParser.Initializer_valueContext):
        pass

    # Enter a parse tree produced by CSharpParser#collection_initializer.
    def enterCollection_initializer(self, ctx: CSharpParser.Collection_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#collection_initializer.
    def exitCollection_initializer(self, ctx: CSharpParser.Collection_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#element_initializer.
    def enterElement_initializer(self, ctx: CSharpParser.Element_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#element_initializer.
    def exitElement_initializer(self, ctx: CSharpParser.Element_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#anonymous_object_initializer.
    def enterAnonymous_object_initializer(self, ctx: CSharpParser.Anonymous_object_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#anonymous_object_initializer.
    def exitAnonymous_object_initializer(self, ctx: CSharpParser.Anonymous_object_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_declarator_list.
    def enterMember_declarator_list(self, ctx: CSharpParser.Member_declarator_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_declarator_list.
    def exitMember_declarator_list(self, ctx: CSharpParser.Member_declarator_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_declarator.
    def enterMember_declarator(self, ctx: CSharpParser.Member_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_declarator.
    def exitMember_declarator(self, ctx: CSharpParser.Member_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#unbound_type_name.
    def enterUnbound_type_name(self, ctx: CSharpParser.Unbound_type_nameContext):
        pass

    # Exit a parse tree produced by CSharpParser#unbound_type_name.
    def exitUnbound_type_name(self, ctx: CSharpParser.Unbound_type_nameContext):
        pass

    # Enter a parse tree produced by CSharpParser#generic_dimension_specifier.
    def enterGeneric_dimension_specifier(self, ctx: CSharpParser.Generic_dimension_specifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#generic_dimension_specifier.
    def exitGeneric_dimension_specifier(self, ctx: CSharpParser.Generic_dimension_specifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#isType.
    def enterIsType(self, ctx: CSharpParser.IsTypeContext):
        pass

    # Exit a parse tree produced by CSharpParser#isType.
    def exitIsType(self, ctx: CSharpParser.IsTypeContext):
        pass

    # Enter a parse tree produced by CSharpParser#isTypePatternArms.
    def enterIsTypePatternArms(self, ctx: CSharpParser.IsTypePatternArmsContext):
        pass

    # Exit a parse tree produced by CSharpParser#isTypePatternArms.
    def exitIsTypePatternArms(self, ctx: CSharpParser.IsTypePatternArmsContext):
        pass

    # Enter a parse tree produced by CSharpParser#isTypePatternArm.
    def enterIsTypePatternArm(self, ctx: CSharpParser.IsTypePatternArmContext):
        pass

    # Exit a parse tree produced by CSharpParser#isTypePatternArm.
    def exitIsTypePatternArm(self, ctx: CSharpParser.IsTypePatternArmContext):
        pass

    # Enter a parse tree produced by CSharpParser#lambda_expression.
    def enterLambda_expression(self, ctx: CSharpParser.Lambda_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#lambda_expression.
    def exitLambda_expression(self, ctx: CSharpParser.Lambda_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#anonymous_function_signature.
    def enterAnonymous_function_signature(self, ctx: CSharpParser.Anonymous_function_signatureContext):
        pass

    # Exit a parse tree produced by CSharpParser#anonymous_function_signature.
    def exitAnonymous_function_signature(self, ctx: CSharpParser.Anonymous_function_signatureContext):
        pass

    # Enter a parse tree produced by CSharpParser#explicit_anonymous_function_parameter_list.
    def enterExplicit_anonymous_function_parameter_list(self, ctx: CSharpParser.Explicit_anonymous_function_parameter_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#explicit_anonymous_function_parameter_list.
    def exitExplicit_anonymous_function_parameter_list(self, ctx: CSharpParser.Explicit_anonymous_function_parameter_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#explicit_anonymous_function_parameter.
    def enterExplicit_anonymous_function_parameter(self, ctx: CSharpParser.Explicit_anonymous_function_parameterContext):
        pass

    # Exit a parse tree produced by CSharpParser#explicit_anonymous_function_parameter.
    def exitExplicit_anonymous_function_parameter(self, ctx: CSharpParser.Explicit_anonymous_function_parameterContext):
        pass

    # Enter a parse tree produced by CSharpParser#implicit_anonymous_function_parameter_list.
    def enterImplicit_anonymous_function_parameter_list(self, ctx: CSharpParser.Implicit_anonymous_function_parameter_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#implicit_anonymous_function_parameter_list.
    def exitImplicit_anonymous_function_parameter_list(self, ctx: CSharpParser.Implicit_anonymous_function_parameter_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#anonymous_function_body.
    def enterAnonymous_function_body(self, ctx: CSharpParser.Anonymous_function_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#anonymous_function_body.
    def exitAnonymous_function_body(self, ctx: CSharpParser.Anonymous_function_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#query_expression.
    def enterQuery_expression(self, ctx: CSharpParser.Query_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#query_expression.
    def exitQuery_expression(self, ctx: CSharpParser.Query_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#from_clause.
    def enterFrom_clause(self, ctx: CSharpParser.From_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#from_clause.
    def exitFrom_clause(self, ctx: CSharpParser.From_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#query_body.
    def enterQuery_body(self, ctx: CSharpParser.Query_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#query_body.
    def exitQuery_body(self, ctx: CSharpParser.Query_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#query_body_clause.
    def enterQuery_body_clause(self, ctx: CSharpParser.Query_body_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#query_body_clause.
    def exitQuery_body_clause(self, ctx: CSharpParser.Query_body_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#let_clause.
    def enterLet_clause(self, ctx: CSharpParser.Let_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#let_clause.
    def exitLet_clause(self, ctx: CSharpParser.Let_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#where_clause.
    def enterWhere_clause(self, ctx: CSharpParser.Where_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#where_clause.
    def exitWhere_clause(self, ctx: CSharpParser.Where_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#combined_join_clause.
    def enterCombined_join_clause(self, ctx: CSharpParser.Combined_join_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#combined_join_clause.
    def exitCombined_join_clause(self, ctx: CSharpParser.Combined_join_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#orderby_clause.
    def enterOrderby_clause(self, ctx: CSharpParser.Orderby_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#orderby_clause.
    def exitOrderby_clause(self, ctx: CSharpParser.Orderby_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#ordering.
    def enterOrdering(self, ctx: CSharpParser.OrderingContext):
        pass

    # Exit a parse tree produced by CSharpParser#ordering.
    def exitOrdering(self, ctx: CSharpParser.OrderingContext):
        pass

    # Enter a parse tree produced by CSharpParser#select_or_group_clause.
    def enterSelect_or_group_clause(self, ctx: CSharpParser.Select_or_group_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#select_or_group_clause.
    def exitSelect_or_group_clause(self, ctx: CSharpParser.Select_or_group_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#query_continuation.
    def enterQuery_continuation(self, ctx: CSharpParser.Query_continuationContext):
        pass

    # Exit a parse tree produced by CSharpParser#query_continuation.
    def exitQuery_continuation(self, ctx: CSharpParser.Query_continuationContext):
        pass

    # Enter a parse tree produced by CSharpParser#statement.
    def enterStatement(self, ctx: CSharpParser.StatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#statement.
    def exitStatement(self, ctx: CSharpParser.StatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#declarationStatement.
    def enterDeclarationStatement(self, ctx: CSharpParser.DeclarationStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#declarationStatement.
    def exitDeclarationStatement(self, ctx: CSharpParser.DeclarationStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_function_declaration.
    def enterLocal_function_declaration(self, ctx: CSharpParser.Local_function_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_function_declaration.
    def exitLocal_function_declaration(self, ctx: CSharpParser.Local_function_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_function_header.
    def enterLocal_function_header(self, ctx: CSharpParser.Local_function_headerContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_function_header.
    def exitLocal_function_header(self, ctx: CSharpParser.Local_function_headerContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_function_modifiers.
    def enterLocal_function_modifiers(self, ctx: CSharpParser.Local_function_modifiersContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_function_modifiers.
    def exitLocal_function_modifiers(self, ctx: CSharpParser.Local_function_modifiersContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_function_body.
    def enterLocal_function_body(self, ctx: CSharpParser.Local_function_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_function_body.
    def exitLocal_function_body(self, ctx: CSharpParser.Local_function_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#labeled_Statement.
    def enterLabeled_Statement(self, ctx: CSharpParser.Labeled_StatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#labeled_Statement.
    def exitLabeled_Statement(self, ctx: CSharpParser.Labeled_StatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#embedded_statement.
    def enterEmbedded_statement(self, ctx: CSharpParser.Embedded_statementContext):
        pass

    # Exit a parse tree produced by CSharpParser#embedded_statement.
    def exitEmbedded_statement(self, ctx: CSharpParser.Embedded_statementContext):
        pass

    # Enter a parse tree produced by CSharpParser#theEmptyStatement.
    def enterTheEmptyStatement(self, ctx: CSharpParser.TheEmptyStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#theEmptyStatement.
    def exitTheEmptyStatement(self, ctx: CSharpParser.TheEmptyStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#expressionStatement.
    def enterExpressionStatement(self, ctx: CSharpParser.ExpressionStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#expressionStatement.
    def exitExpressionStatement(self, ctx: CSharpParser.ExpressionStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#ifStatement.
    def enterIfStatement(self, ctx: CSharpParser.IfStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#ifStatement.
    def exitIfStatement(self, ctx: CSharpParser.IfStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#switchStatement.
    def enterSwitchStatement(self, ctx: CSharpParser.SwitchStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#switchStatement.
    def exitSwitchStatement(self, ctx: CSharpParser.SwitchStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#whileStatement.
    def enterWhileStatement(self, ctx: CSharpParser.WhileStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#whileStatement.
    def exitWhileStatement(self, ctx: CSharpParser.WhileStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#doStatement.
    def enterDoStatement(self, ctx: CSharpParser.DoStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#doStatement.
    def exitDoStatement(self, ctx: CSharpParser.DoStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#forStatement.
    def enterForStatement(self, ctx: CSharpParser.ForStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#forStatement.
    def exitForStatement(self, ctx: CSharpParser.ForStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#foreachStatement.
    def enterForeachStatement(self, ctx: CSharpParser.ForeachStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#foreachStatement.
    def exitForeachStatement(self, ctx: CSharpParser.ForeachStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#breakStatement.
    def enterBreakStatement(self, ctx: CSharpParser.BreakStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#breakStatement.
    def exitBreakStatement(self, ctx: CSharpParser.BreakStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#continueStatement.
    def enterContinueStatement(self, ctx: CSharpParser.ContinueStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#continueStatement.
    def exitContinueStatement(self, ctx: CSharpParser.ContinueStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#gotoStatement.
    def enterGotoStatement(self, ctx: CSharpParser.GotoStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#gotoStatement.
    def exitGotoStatement(self, ctx: CSharpParser.GotoStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#returnStatement.
    def enterReturnStatement(self, ctx: CSharpParser.ReturnStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#returnStatement.
    def exitReturnStatement(self, ctx: CSharpParser.ReturnStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#throwStatement.
    def enterThrowStatement(self, ctx: CSharpParser.ThrowStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#throwStatement.
    def exitThrowStatement(self, ctx: CSharpParser.ThrowStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#tryStatement.
    def enterTryStatement(self, ctx: CSharpParser.TryStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#tryStatement.
    def exitTryStatement(self, ctx: CSharpParser.TryStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#checkedStatement.
    def enterCheckedStatement(self, ctx: CSharpParser.CheckedStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#checkedStatement.
    def exitCheckedStatement(self, ctx: CSharpParser.CheckedStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#uncheckedStatement.
    def enterUncheckedStatement(self, ctx: CSharpParser.UncheckedStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#uncheckedStatement.
    def exitUncheckedStatement(self, ctx: CSharpParser.UncheckedStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#lockStatement.
    def enterLockStatement(self, ctx: CSharpParser.LockStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#lockStatement.
    def exitLockStatement(self, ctx: CSharpParser.LockStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#usingStatement.
    def enterUsingStatement(self, ctx: CSharpParser.UsingStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#usingStatement.
    def exitUsingStatement(self, ctx: CSharpParser.UsingStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#yieldStatement.
    def enterYieldStatement(self, ctx: CSharpParser.YieldStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#yieldStatement.
    def exitYieldStatement(self, ctx: CSharpParser.YieldStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#unsafeStatement.
    def enterUnsafeStatement(self, ctx: CSharpParser.UnsafeStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#unsafeStatement.
    def exitUnsafeStatement(self, ctx: CSharpParser.UnsafeStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixedStatement.
    def enterFixedStatement(self, ctx: CSharpParser.FixedStatementContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixedStatement.
    def exitFixedStatement(self, ctx: CSharpParser.FixedStatementContext):
        pass

    # Enter a parse tree produced by CSharpParser#block.
    def enterBlock(self, ctx: CSharpParser.BlockContext):
        pass

    # Exit a parse tree produced by CSharpParser#block.
    def exitBlock(self, ctx: CSharpParser.BlockContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_variable_declaration.
    def enterLocal_variable_declaration(self, ctx: CSharpParser.Local_variable_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_variable_declaration.
    def exitLocal_variable_declaration(self, ctx: CSharpParser.Local_variable_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_variable_type.
    def enterLocal_variable_type(self, ctx: CSharpParser.Local_variable_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_variable_type.
    def exitLocal_variable_type(self, ctx: CSharpParser.Local_variable_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_variable_declarator.
    def enterLocal_variable_declarator(self, ctx: CSharpParser.Local_variable_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_variable_declarator.
    def exitLocal_variable_declarator(self, ctx: CSharpParser.Local_variable_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_variable_initializer.
    def enterLocal_variable_initializer(self, ctx: CSharpParser.Local_variable_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_variable_initializer.
    def exitLocal_variable_initializer(self, ctx: CSharpParser.Local_variable_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#local_constant_declaration.
    def enterLocal_constant_declaration(self, ctx: CSharpParser.Local_constant_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#local_constant_declaration.
    def exitLocal_constant_declaration(self, ctx: CSharpParser.Local_constant_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#if_body.
    def enterIf_body(self, ctx: CSharpParser.If_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#if_body.
    def exitIf_body(self, ctx: CSharpParser.If_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#switch_section.
    def enterSwitch_section(self, ctx: CSharpParser.Switch_sectionContext):
        pass

    # Exit a parse tree produced by CSharpParser#switch_section.
    def exitSwitch_section(self, ctx: CSharpParser.Switch_sectionContext):
        pass

    # Enter a parse tree produced by CSharpParser#switch_label.
    def enterSwitch_label(self, ctx: CSharpParser.Switch_labelContext):
        pass

    # Exit a parse tree produced by CSharpParser#switch_label.
    def exitSwitch_label(self, ctx: CSharpParser.Switch_labelContext):
        pass

    # Enter a parse tree produced by CSharpParser#case_guard.
    def enterCase_guard(self, ctx: CSharpParser.Case_guardContext):
        pass

    # Exit a parse tree produced by CSharpParser#case_guard.
    def exitCase_guard(self, ctx: CSharpParser.Case_guardContext):
        pass

    # Enter a parse tree produced by CSharpParser#statement_list.
    def enterStatement_list(self, ctx: CSharpParser.Statement_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#statement_list.
    def exitStatement_list(self, ctx: CSharpParser.Statement_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#for_initializer.
    def enterFor_initializer(self, ctx: CSharpParser.For_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#for_initializer.
    def exitFor_initializer(self, ctx: CSharpParser.For_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#for_iterator.
    def enterFor_iterator(self, ctx: CSharpParser.For_iteratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#for_iterator.
    def exitFor_iterator(self, ctx: CSharpParser.For_iteratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#catch_clauses.
    def enterCatch_clauses(self, ctx: CSharpParser.Catch_clausesContext):
        pass

    # Exit a parse tree produced by CSharpParser#catch_clauses.
    def exitCatch_clauses(self, ctx: CSharpParser.Catch_clausesContext):
        pass

    # Enter a parse tree produced by CSharpParser#specific_catch_clause.
    def enterSpecific_catch_clause(self, ctx: CSharpParser.Specific_catch_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#specific_catch_clause.
    def exitSpecific_catch_clause(self, ctx: CSharpParser.Specific_catch_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#general_catch_clause.
    def enterGeneral_catch_clause(self, ctx: CSharpParser.General_catch_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#general_catch_clause.
    def exitGeneral_catch_clause(self, ctx: CSharpParser.General_catch_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#exception_filter.
    def enterException_filter(self, ctx: CSharpParser.Exception_filterContext):
        pass

    # Exit a parse tree produced by CSharpParser#exception_filter.
    def exitException_filter(self, ctx: CSharpParser.Exception_filterContext):
        pass

    # Enter a parse tree produced by CSharpParser#finally_clause.
    def enterFinally_clause(self, ctx: CSharpParser.Finally_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#finally_clause.
    def exitFinally_clause(self, ctx: CSharpParser.Finally_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#resource_acquisition.
    def enterResource_acquisition(self, ctx: CSharpParser.Resource_acquisitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#resource_acquisition.
    def exitResource_acquisition(self, ctx: CSharpParser.Resource_acquisitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#namespace_declaration.
    def enterNamespace_declaration(self, ctx: CSharpParser.Namespace_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#namespace_declaration.
    def exitNamespace_declaration(self, ctx: CSharpParser.Namespace_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#qualified_identifier.
    def enterQualified_identifier(self, ctx: CSharpParser.Qualified_identifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#qualified_identifier.
    def exitQualified_identifier(self, ctx: CSharpParser.Qualified_identifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#namespace_body.
    def enterNamespace_body(self, ctx: CSharpParser.Namespace_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#namespace_body.
    def exitNamespace_body(self, ctx: CSharpParser.Namespace_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#extern_alias_directives.
    def enterExtern_alias_directives(self, ctx: CSharpParser.Extern_alias_directivesContext):
        pass

    # Exit a parse tree produced by CSharpParser#extern_alias_directives.
    def exitExtern_alias_directives(self, ctx: CSharpParser.Extern_alias_directivesContext):
        pass

    # Enter a parse tree produced by CSharpParser#extern_alias_directive.
    def enterExtern_alias_directive(self, ctx: CSharpParser.Extern_alias_directiveContext):
        pass

    # Exit a parse tree produced by CSharpParser#extern_alias_directive.
    def exitExtern_alias_directive(self, ctx: CSharpParser.Extern_alias_directiveContext):
        pass

    # Enter a parse tree produced by CSharpParser#using_directives.
    def enterUsing_directives(self, ctx: CSharpParser.Using_directivesContext):
        pass

    # Exit a parse tree produced by CSharpParser#using_directives.
    def exitUsing_directives(self, ctx: CSharpParser.Using_directivesContext):
        pass

    # Enter a parse tree produced by CSharpParser#usingAliasDirective.
    def enterUsingAliasDirective(self, ctx: CSharpParser.UsingAliasDirectiveContext):
        pass

    # Exit a parse tree produced by CSharpParser#usingAliasDirective.
    def exitUsingAliasDirective(self, ctx: CSharpParser.UsingAliasDirectiveContext):
        pass

    # Enter a parse tree produced by CSharpParser#usingNamespaceDirective.
    def enterUsingNamespaceDirective(self, ctx: CSharpParser.UsingNamespaceDirectiveContext):
        pass

    # Exit a parse tree produced by CSharpParser#usingNamespaceDirective.
    def exitUsingNamespaceDirective(self, ctx: CSharpParser.UsingNamespaceDirectiveContext):
        pass

    # Enter a parse tree produced by CSharpParser#usingStaticDirective.
    def enterUsingStaticDirective(self, ctx: CSharpParser.UsingStaticDirectiveContext):
        pass

    # Exit a parse tree produced by CSharpParser#usingStaticDirective.
    def exitUsingStaticDirective(self, ctx: CSharpParser.UsingStaticDirectiveContext):
        pass

    # Enter a parse tree produced by CSharpParser#namespace_member_declarations.
    def enterNamespace_member_declarations(self, ctx: CSharpParser.Namespace_member_declarationsContext):
        pass

    # Exit a parse tree produced by CSharpParser#namespace_member_declarations.
    def exitNamespace_member_declarations(self, ctx: CSharpParser.Namespace_member_declarationsContext):
        pass

    # Enter a parse tree produced by CSharpParser#namespace_member_declaration.
    def enterNamespace_member_declaration(self, ctx: CSharpParser.Namespace_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#namespace_member_declaration.
    def exitNamespace_member_declaration(self, ctx: CSharpParser.Namespace_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_declaration.
    def enterType_declaration(self, ctx: CSharpParser.Type_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_declaration.
    def exitType_declaration(self, ctx: CSharpParser.Type_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#qualified_alias_member.
    def enterQualified_alias_member(self, ctx: CSharpParser.Qualified_alias_memberContext):
        pass

    # Exit a parse tree produced by CSharpParser#qualified_alias_member.
    def exitQualified_alias_member(self, ctx: CSharpParser.Qualified_alias_memberContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_parameter_list.
    def enterType_parameter_list(self, ctx: CSharpParser.Type_parameter_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_parameter_list.
    def exitType_parameter_list(self, ctx: CSharpParser.Type_parameter_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_parameter.
    def enterType_parameter(self, ctx: CSharpParser.Type_parameterContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_parameter.
    def exitType_parameter(self, ctx: CSharpParser.Type_parameterContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_base.
    def enterClass_base(self, ctx: CSharpParser.Class_baseContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_base.
    def exitClass_base(self, ctx: CSharpParser.Class_baseContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_type_list.
    def enterInterface_type_list(self, ctx: CSharpParser.Interface_type_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_type_list.
    def exitInterface_type_list(self, ctx: CSharpParser.Interface_type_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_parameter_constraints_clauses.
    def enterType_parameter_constraints_clauses(self, ctx: CSharpParser.Type_parameter_constraints_clausesContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_parameter_constraints_clauses.
    def exitType_parameter_constraints_clauses(self, ctx: CSharpParser.Type_parameter_constraints_clausesContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_parameter_constraints_clause.
    def enterType_parameter_constraints_clause(self, ctx: CSharpParser.Type_parameter_constraints_clauseContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_parameter_constraints_clause.
    def exitType_parameter_constraints_clause(self, ctx: CSharpParser.Type_parameter_constraints_clauseContext):
        pass

    # Enter a parse tree produced by CSharpParser#type_parameter_constraints.
    def enterType_parameter_constraints(self, ctx: CSharpParser.Type_parameter_constraintsContext):
        pass

    # Exit a parse tree produced by CSharpParser#type_parameter_constraints.
    def exitType_parameter_constraints(self, ctx: CSharpParser.Type_parameter_constraintsContext):
        pass

    # Enter a parse tree produced by CSharpParser#primary_constraint.
    def enterPrimary_constraint(self, ctx: CSharpParser.Primary_constraintContext):
        pass

    # Exit a parse tree produced by CSharpParser#primary_constraint.
    def exitPrimary_constraint(self, ctx: CSharpParser.Primary_constraintContext):
        pass

    # Enter a parse tree produced by CSharpParser#secondary_constraints.
    def enterSecondary_constraints(self, ctx: CSharpParser.Secondary_constraintsContext):
        pass

    # Exit a parse tree produced by CSharpParser#secondary_constraints.
    def exitSecondary_constraints(self, ctx: CSharpParser.Secondary_constraintsContext):
        pass

    # Enter a parse tree produced by CSharpParser#constructor_constraint.
    def enterConstructor_constraint(self, ctx: CSharpParser.Constructor_constraintContext):
        pass

    # Exit a parse tree produced by CSharpParser#constructor_constraint.
    def exitConstructor_constraint(self, ctx: CSharpParser.Constructor_constraintContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_body.
    def enterClass_body(self, ctx: CSharpParser.Class_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_body.
    def exitClass_body(self, ctx: CSharpParser.Class_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_member_declarations.
    def enterClass_member_declarations(self, ctx: CSharpParser.Class_member_declarationsContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_member_declarations.
    def exitClass_member_declarations(self, ctx: CSharpParser.Class_member_declarationsContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_member_declaration.
    def enterClass_member_declaration(self, ctx: CSharpParser.Class_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_member_declaration.
    def exitClass_member_declaration(self, ctx: CSharpParser.Class_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#all_member_modifiers.
    def enterAll_member_modifiers(self, ctx: CSharpParser.All_member_modifiersContext):
        pass

    # Exit a parse tree produced by CSharpParser#all_member_modifiers.
    def exitAll_member_modifiers(self, ctx: CSharpParser.All_member_modifiersContext):
        pass

    # Enter a parse tree produced by CSharpParser#all_member_modifier.
    def enterAll_member_modifier(self, ctx: CSharpParser.All_member_modifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#all_member_modifier.
    def exitAll_member_modifier(self, ctx: CSharpParser.All_member_modifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#common_member_declaration.
    def enterCommon_member_declaration(self, ctx: CSharpParser.Common_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#common_member_declaration.
    def exitCommon_member_declaration(self, ctx: CSharpParser.Common_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#typed_member_declaration.
    def enterTyped_member_declaration(self, ctx: CSharpParser.Typed_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#typed_member_declaration.
    def exitTyped_member_declaration(self, ctx: CSharpParser.Typed_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#constant_declarators.
    def enterConstant_declarators(self, ctx: CSharpParser.Constant_declaratorsContext):
        pass

    # Exit a parse tree produced by CSharpParser#constant_declarators.
    def exitConstant_declarators(self, ctx: CSharpParser.Constant_declaratorsContext):
        pass

    # Enter a parse tree produced by CSharpParser#constant_declarator.
    def enterConstant_declarator(self, ctx: CSharpParser.Constant_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#constant_declarator.
    def exitConstant_declarator(self, ctx: CSharpParser.Constant_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#variable_declarators.
    def enterVariable_declarators(self, ctx: CSharpParser.Variable_declaratorsContext):
        pass

    # Exit a parse tree produced by CSharpParser#variable_declarators.
    def exitVariable_declarators(self, ctx: CSharpParser.Variable_declaratorsContext):
        pass

    # Enter a parse tree produced by CSharpParser#variable_declarator.
    def enterVariable_declarator(self, ctx: CSharpParser.Variable_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#variable_declarator.
    def exitVariable_declarator(self, ctx: CSharpParser.Variable_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#variable_initializer.
    def enterVariable_initializer(self, ctx: CSharpParser.Variable_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#variable_initializer.
    def exitVariable_initializer(self, ctx: CSharpParser.Variable_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#return_type.
    def enterReturn_type(self, ctx: CSharpParser.Return_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#return_type.
    def exitReturn_type(self, ctx: CSharpParser.Return_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#member_name.
    def enterMember_name(self, ctx: CSharpParser.Member_nameContext):
        pass

    # Exit a parse tree produced by CSharpParser#member_name.
    def exitMember_name(self, ctx: CSharpParser.Member_nameContext):
        pass

    # Enter a parse tree produced by CSharpParser#method_body.
    def enterMethod_body(self, ctx: CSharpParser.Method_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#method_body.
    def exitMethod_body(self, ctx: CSharpParser.Method_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#formal_parameter_list.
    def enterFormal_parameter_list(self, ctx: CSharpParser.Formal_parameter_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#formal_parameter_list.
    def exitFormal_parameter_list(self, ctx: CSharpParser.Formal_parameter_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_parameters.
    def enterFixed_parameters(self, ctx: CSharpParser.Fixed_parametersContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_parameters.
    def exitFixed_parameters(self, ctx: CSharpParser.Fixed_parametersContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_parameter.
    def enterFixed_parameter(self, ctx: CSharpParser.Fixed_parameterContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_parameter.
    def exitFixed_parameter(self, ctx: CSharpParser.Fixed_parameterContext):
        pass

    # Enter a parse tree produced by CSharpParser#parameter_modifier.
    def enterParameter_modifier(self, ctx: CSharpParser.Parameter_modifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#parameter_modifier.
    def exitParameter_modifier(self, ctx: CSharpParser.Parameter_modifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#parameter_array.
    def enterParameter_array(self, ctx: CSharpParser.Parameter_arrayContext):
        pass

    # Exit a parse tree produced by CSharpParser#parameter_array.
    def exitParameter_array(self, ctx: CSharpParser.Parameter_arrayContext):
        pass

    # Enter a parse tree produced by CSharpParser#accessor_declarations.
    def enterAccessor_declarations(self, ctx: CSharpParser.Accessor_declarationsContext):
        pass

    # Exit a parse tree produced by CSharpParser#accessor_declarations.
    def exitAccessor_declarations(self, ctx: CSharpParser.Accessor_declarationsContext):
        pass

    # Enter a parse tree produced by CSharpParser#get_accessor_declaration.
    def enterGet_accessor_declaration(self, ctx: CSharpParser.Get_accessor_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#get_accessor_declaration.
    def exitGet_accessor_declaration(self, ctx: CSharpParser.Get_accessor_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#set_accessor_declaration.
    def enterSet_accessor_declaration(self, ctx: CSharpParser.Set_accessor_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#set_accessor_declaration.
    def exitSet_accessor_declaration(self, ctx: CSharpParser.Set_accessor_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#accessor_modifier.
    def enterAccessor_modifier(self, ctx: CSharpParser.Accessor_modifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#accessor_modifier.
    def exitAccessor_modifier(self, ctx: CSharpParser.Accessor_modifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#accessor_body.
    def enterAccessor_body(self, ctx: CSharpParser.Accessor_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#accessor_body.
    def exitAccessor_body(self, ctx: CSharpParser.Accessor_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#event_accessor_declarations.
    def enterEvent_accessor_declarations(self, ctx: CSharpParser.Event_accessor_declarationsContext):
        pass

    # Exit a parse tree produced by CSharpParser#event_accessor_declarations.
    def exitEvent_accessor_declarations(self, ctx: CSharpParser.Event_accessor_declarationsContext):
        pass

    # Enter a parse tree produced by CSharpParser#add_accessor_declaration.
    def enterAdd_accessor_declaration(self, ctx: CSharpParser.Add_accessor_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#add_accessor_declaration.
    def exitAdd_accessor_declaration(self, ctx: CSharpParser.Add_accessor_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#remove_accessor_declaration.
    def enterRemove_accessor_declaration(self, ctx: CSharpParser.Remove_accessor_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#remove_accessor_declaration.
    def exitRemove_accessor_declaration(self, ctx: CSharpParser.Remove_accessor_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#overloadable_operator.
    def enterOverloadable_operator(self, ctx: CSharpParser.Overloadable_operatorContext):
        pass

    # Exit a parse tree produced by CSharpParser#overloadable_operator.
    def exitOverloadable_operator(self, ctx: CSharpParser.Overloadable_operatorContext):
        pass

    # Enter a parse tree produced by CSharpParser#conversion_operator_declarator.
    def enterConversion_operator_declarator(self, ctx: CSharpParser.Conversion_operator_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#conversion_operator_declarator.
    def exitConversion_operator_declarator(self, ctx: CSharpParser.Conversion_operator_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#constructor_initializer.
    def enterConstructor_initializer(self, ctx: CSharpParser.Constructor_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#constructor_initializer.
    def exitConstructor_initializer(self, ctx: CSharpParser.Constructor_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#body.
    def enterBody(self, ctx: CSharpParser.BodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#body.
    def exitBody(self, ctx: CSharpParser.BodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#struct_interfaces.
    def enterStruct_interfaces(self, ctx: CSharpParser.Struct_interfacesContext):
        pass

    # Exit a parse tree produced by CSharpParser#struct_interfaces.
    def exitStruct_interfaces(self, ctx: CSharpParser.Struct_interfacesContext):
        pass

    # Enter a parse tree produced by CSharpParser#struct_body.
    def enterStruct_body(self, ctx: CSharpParser.Struct_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#struct_body.
    def exitStruct_body(self, ctx: CSharpParser.Struct_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#struct_member_declaration.
    def enterStruct_member_declaration(self, ctx: CSharpParser.Struct_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#struct_member_declaration.
    def exitStruct_member_declaration(self, ctx: CSharpParser.Struct_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#array_type.
    def enterArray_type(self, ctx: CSharpParser.Array_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#array_type.
    def exitArray_type(self, ctx: CSharpParser.Array_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#rank_specifier.
    def enterRank_specifier(self, ctx: CSharpParser.Rank_specifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#rank_specifier.
    def exitRank_specifier(self, ctx: CSharpParser.Rank_specifierContext):
        pass

    # Enter a parse tree produced by CSharpParser#array_initializer.
    def enterArray_initializer(self, ctx: CSharpParser.Array_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#array_initializer.
    def exitArray_initializer(self, ctx: CSharpParser.Array_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#variant_type_parameter_list.
    def enterVariant_type_parameter_list(self, ctx: CSharpParser.Variant_type_parameter_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#variant_type_parameter_list.
    def exitVariant_type_parameter_list(self, ctx: CSharpParser.Variant_type_parameter_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#variant_type_parameter.
    def enterVariant_type_parameter(self, ctx: CSharpParser.Variant_type_parameterContext):
        pass

    # Exit a parse tree produced by CSharpParser#variant_type_parameter.
    def exitVariant_type_parameter(self, ctx: CSharpParser.Variant_type_parameterContext):
        pass

    # Enter a parse tree produced by CSharpParser#variance_annotation.
    def enterVariance_annotation(self, ctx: CSharpParser.Variance_annotationContext):
        pass

    # Exit a parse tree produced by CSharpParser#variance_annotation.
    def exitVariance_annotation(self, ctx: CSharpParser.Variance_annotationContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_base.
    def enterInterface_base(self, ctx: CSharpParser.Interface_baseContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_base.
    def exitInterface_base(self, ctx: CSharpParser.Interface_baseContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_body.
    def enterInterface_body(self, ctx: CSharpParser.Interface_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_body.
    def exitInterface_body(self, ctx: CSharpParser.Interface_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_member_declaration.
    def enterInterface_member_declaration(self, ctx: CSharpParser.Interface_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_member_declaration.
    def exitInterface_member_declaration(self, ctx: CSharpParser.Interface_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_accessors.
    def enterInterface_accessors(self, ctx: CSharpParser.Interface_accessorsContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_accessors.
    def exitInterface_accessors(self, ctx: CSharpParser.Interface_accessorsContext):
        pass

    # Enter a parse tree produced by CSharpParser#enum_base.
    def enterEnum_base(self, ctx: CSharpParser.Enum_baseContext):
        pass

    # Exit a parse tree produced by CSharpParser#enum_base.
    def exitEnum_base(self, ctx: CSharpParser.Enum_baseContext):
        pass

    # Enter a parse tree produced by CSharpParser#enum_body.
    def enterEnum_body(self, ctx: CSharpParser.Enum_bodyContext):
        pass

    # Exit a parse tree produced by CSharpParser#enum_body.
    def exitEnum_body(self, ctx: CSharpParser.Enum_bodyContext):
        pass

    # Enter a parse tree produced by CSharpParser#enum_member_declaration.
    def enterEnum_member_declaration(self, ctx: CSharpParser.Enum_member_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#enum_member_declaration.
    def exitEnum_member_declaration(self, ctx: CSharpParser.Enum_member_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#global_attribute_section.
    def enterGlobal_attribute_section(self, ctx: CSharpParser.Global_attribute_sectionContext):
        pass

    # Exit a parse tree produced by CSharpParser#global_attribute_section.
    def exitGlobal_attribute_section(self, ctx: CSharpParser.Global_attribute_sectionContext):
        pass

    # Enter a parse tree produced by CSharpParser#global_attribute_target.
    def enterGlobal_attribute_target(self, ctx: CSharpParser.Global_attribute_targetContext):
        pass

    # Exit a parse tree produced by CSharpParser#global_attribute_target.
    def exitGlobal_attribute_target(self, ctx: CSharpParser.Global_attribute_targetContext):
        pass

    # Enter a parse tree produced by CSharpParser#attributes.
    def enterAttributes(self, ctx: CSharpParser.AttributesContext):
        pass

    # Exit a parse tree produced by CSharpParser#attributes.
    def exitAttributes(self, ctx: CSharpParser.AttributesContext):
        pass

    # Enter a parse tree produced by CSharpParser#attribute_section.
    def enterAttribute_section(self, ctx: CSharpParser.Attribute_sectionContext):
        pass

    # Exit a parse tree produced by CSharpParser#attribute_section.
    def exitAttribute_section(self, ctx: CSharpParser.Attribute_sectionContext):
        pass

    # Enter a parse tree produced by CSharpParser#attribute_target.
    def enterAttribute_target(self, ctx: CSharpParser.Attribute_targetContext):
        pass

    # Exit a parse tree produced by CSharpParser#attribute_target.
    def exitAttribute_target(self, ctx: CSharpParser.Attribute_targetContext):
        pass

    # Enter a parse tree produced by CSharpParser#attribute_list.
    def enterAttribute_list(self, ctx: CSharpParser.Attribute_listContext):
        pass

    # Exit a parse tree produced by CSharpParser#attribute_list.
    def exitAttribute_list(self, ctx: CSharpParser.Attribute_listContext):
        pass

    # Enter a parse tree produced by CSharpParser#attribute.
    def enterAttribute(self, ctx: CSharpParser.AttributeContext):
        pass

    # Exit a parse tree produced by CSharpParser#attribute.
    def exitAttribute(self, ctx: CSharpParser.AttributeContext):
        pass

    # Enter a parse tree produced by CSharpParser#attribute_argument.
    def enterAttribute_argument(self, ctx: CSharpParser.Attribute_argumentContext):
        pass

    # Exit a parse tree produced by CSharpParser#attribute_argument.
    def exitAttribute_argument(self, ctx: CSharpParser.Attribute_argumentContext):
        pass

    # Enter a parse tree produced by CSharpParser#pointer_type.
    def enterPointer_type(self, ctx: CSharpParser.Pointer_typeContext):
        pass

    # Exit a parse tree produced by CSharpParser#pointer_type.
    def exitPointer_type(self, ctx: CSharpParser.Pointer_typeContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_pointer_declarators.
    def enterFixed_pointer_declarators(self, ctx: CSharpParser.Fixed_pointer_declaratorsContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_pointer_declarators.
    def exitFixed_pointer_declarators(self, ctx: CSharpParser.Fixed_pointer_declaratorsContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_pointer_declarator.
    def enterFixed_pointer_declarator(self, ctx: CSharpParser.Fixed_pointer_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_pointer_declarator.
    def exitFixed_pointer_declarator(self, ctx: CSharpParser.Fixed_pointer_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_pointer_initializer.
    def enterFixed_pointer_initializer(self, ctx: CSharpParser.Fixed_pointer_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_pointer_initializer.
    def exitFixed_pointer_initializer(self, ctx: CSharpParser.Fixed_pointer_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#fixed_size_buffer_declarator.
    def enterFixed_size_buffer_declarator(self, ctx: CSharpParser.Fixed_size_buffer_declaratorContext):
        pass

    # Exit a parse tree produced by CSharpParser#fixed_size_buffer_declarator.
    def exitFixed_size_buffer_declarator(self, ctx: CSharpParser.Fixed_size_buffer_declaratorContext):
        pass

    # Enter a parse tree produced by CSharpParser#stackalloc_initializer.
    def enterStackalloc_initializer(self, ctx: CSharpParser.Stackalloc_initializerContext):
        pass

    # Exit a parse tree produced by CSharpParser#stackalloc_initializer.
    def exitStackalloc_initializer(self, ctx: CSharpParser.Stackalloc_initializerContext):
        pass

    # Enter a parse tree produced by CSharpParser#right_arrow.
    def enterRight_arrow(self, ctx: CSharpParser.Right_arrowContext):
        pass

    # Exit a parse tree produced by CSharpParser#right_arrow.
    def exitRight_arrow(self, ctx: CSharpParser.Right_arrowContext):
        pass

    # Enter a parse tree produced by CSharpParser#right_shift.
    def enterRight_shift(self, ctx: CSharpParser.Right_shiftContext):
        pass

    # Exit a parse tree produced by CSharpParser#right_shift.
    def exitRight_shift(self, ctx: CSharpParser.Right_shiftContext):
        pass

    # Enter a parse tree produced by CSharpParser#right_shift_assignment.
    def enterRight_shift_assignment(self, ctx: CSharpParser.Right_shift_assignmentContext):
        pass

    # Exit a parse tree produced by CSharpParser#right_shift_assignment.
    def exitRight_shift_assignment(self, ctx: CSharpParser.Right_shift_assignmentContext):
        pass

    # Enter a parse tree produced by CSharpParser#literal.
    def enterLiteral(self, ctx: CSharpParser.LiteralContext):
        pass

    # Exit a parse tree produced by CSharpParser#literal.
    def exitLiteral(self, ctx: CSharpParser.LiteralContext):
        pass

    # Enter a parse tree produced by CSharpParser#boolean_literal.
    def enterBoolean_literal(self, ctx: CSharpParser.Boolean_literalContext):
        pass

    # Exit a parse tree produced by CSharpParser#boolean_literal.
    def exitBoolean_literal(self, ctx: CSharpParser.Boolean_literalContext):
        pass

    # Enter a parse tree produced by CSharpParser#string_literal.
    def enterString_literal(self, ctx: CSharpParser.String_literalContext):
        pass

    # Exit a parse tree produced by CSharpParser#string_literal.
    def exitString_literal(self, ctx: CSharpParser.String_literalContext):
        pass

    # Enter a parse tree produced by CSharpParser#interpolated_regular_string.
    def enterInterpolated_regular_string(self, ctx: CSharpParser.Interpolated_regular_stringContext):
        pass

    # Exit a parse tree produced by CSharpParser#interpolated_regular_string.
    def exitInterpolated_regular_string(self, ctx: CSharpParser.Interpolated_regular_stringContext):
        pass

    # Enter a parse tree produced by CSharpParser#interpolated_verbatium_string.
    def enterInterpolated_verbatium_string(self, ctx: CSharpParser.Interpolated_verbatium_stringContext):
        pass

    # Exit a parse tree produced by CSharpParser#interpolated_verbatium_string.
    def exitInterpolated_verbatium_string(self, ctx: CSharpParser.Interpolated_verbatium_stringContext):
        pass

    # Enter a parse tree produced by CSharpParser#interpolated_regular_string_part.
    def enterInterpolated_regular_string_part(self, ctx: CSharpParser.Interpolated_regular_string_partContext):
        pass

    # Exit a parse tree produced by CSharpParser#interpolated_regular_string_part.
    def exitInterpolated_regular_string_part(self, ctx: CSharpParser.Interpolated_regular_string_partContext):
        pass

    # Enter a parse tree produced by CSharpParser#interpolated_verbatium_string_part.
    def enterInterpolated_verbatium_string_part(self, ctx: CSharpParser.Interpolated_verbatium_string_partContext):
        pass

    # Exit a parse tree produced by CSharpParser#interpolated_verbatium_string_part.
    def exitInterpolated_verbatium_string_part(self, ctx: CSharpParser.Interpolated_verbatium_string_partContext):
        pass

    # Enter a parse tree produced by CSharpParser#interpolated_string_expression.
    def enterInterpolated_string_expression(self, ctx: CSharpParser.Interpolated_string_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#interpolated_string_expression.
    def exitInterpolated_string_expression(self, ctx: CSharpParser.Interpolated_string_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#keyword.
    def enterKeyword(self, ctx: CSharpParser.KeywordContext):
        pass

    # Exit a parse tree produced by CSharpParser#keyword.
    def exitKeyword(self, ctx: CSharpParser.KeywordContext):
        pass

    # Enter a parse tree produced by CSharpParser#class_definition.
    def enterClass_definition(self, ctx: CSharpParser.Class_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#class_definition.
    def exitClass_definition(self, ctx: CSharpParser.Class_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#struct_definition.
    def enterStruct_definition(self, ctx: CSharpParser.Struct_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#struct_definition.
    def exitStruct_definition(self, ctx: CSharpParser.Struct_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#interface_definition.
    def enterInterface_definition(self, ctx: CSharpParser.Interface_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#interface_definition.
    def exitInterface_definition(self, ctx: CSharpParser.Interface_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#enum_definition.
    def enterEnum_definition(self, ctx: CSharpParser.Enum_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#enum_definition.
    def exitEnum_definition(self, ctx: CSharpParser.Enum_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#delegate_definition.
    def enterDelegate_definition(self, ctx: CSharpParser.Delegate_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#delegate_definition.
    def exitDelegate_definition(self, ctx: CSharpParser.Delegate_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#event_declaration.
    def enterEvent_declaration(self, ctx: CSharpParser.Event_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#event_declaration.
    def exitEvent_declaration(self, ctx: CSharpParser.Event_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#field_declaration.
    def enterField_declaration(self, ctx: CSharpParser.Field_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#field_declaration.
    def exitField_declaration(self, ctx: CSharpParser.Field_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#property_declaration.
    def enterProperty_declaration(self, ctx: CSharpParser.Property_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#property_declaration.
    def exitProperty_declaration(self, ctx: CSharpParser.Property_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#constant_declaration.
    def enterConstant_declaration(self, ctx: CSharpParser.Constant_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#constant_declaration.
    def exitConstant_declaration(self, ctx: CSharpParser.Constant_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#indexer_declaration.
    def enterIndexer_declaration(self, ctx: CSharpParser.Indexer_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#indexer_declaration.
    def exitIndexer_declaration(self, ctx: CSharpParser.Indexer_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#destructor_definition.
    def enterDestructor_definition(self, ctx: CSharpParser.Destructor_definitionContext):
        pass

    # Exit a parse tree produced by CSharpParser#destructor_definition.
    def exitDestructor_definition(self, ctx: CSharpParser.Destructor_definitionContext):
        pass

    # Enter a parse tree produced by CSharpParser#constructor_declaration.
    def enterConstructor_declaration(self, ctx: CSharpParser.Constructor_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#constructor_declaration.
    def exitConstructor_declaration(self, ctx: CSharpParser.Constructor_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#method_declaration.
    def enterMethod_declaration(self, ctx: CSharpParser.Method_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#method_declaration.
    def exitMethod_declaration(self, ctx: CSharpParser.Method_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#method_member_name.
    def enterMethod_member_name(self, ctx: CSharpParser.Method_member_nameContext):
        pass

    # Exit a parse tree produced by CSharpParser#method_member_name.
    def exitMethod_member_name(self, ctx: CSharpParser.Method_member_nameContext):
        pass

    # Enter a parse tree produced by CSharpParser#operator_declaration.
    def enterOperator_declaration(self, ctx: CSharpParser.Operator_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#operator_declaration.
    def exitOperator_declaration(self, ctx: CSharpParser.Operator_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#arg_declaration.
    def enterArg_declaration(self, ctx: CSharpParser.Arg_declarationContext):
        pass

    # Exit a parse tree produced by CSharpParser#arg_declaration.
    def exitArg_declaration(self, ctx: CSharpParser.Arg_declarationContext):
        pass

    # Enter a parse tree produced by CSharpParser#method_invocation.
    def enterMethod_invocation(self, ctx: CSharpParser.Method_invocationContext):
        pass

    # Exit a parse tree produced by CSharpParser#method_invocation.
    def exitMethod_invocation(self, ctx: CSharpParser.Method_invocationContext):
        pass

    # Enter a parse tree produced by CSharpParser#object_creation_expression.
    def enterObject_creation_expression(self, ctx: CSharpParser.Object_creation_expressionContext):
        pass

    # Exit a parse tree produced by CSharpParser#object_creation_expression.
    def exitObject_creation_expression(self, ctx: CSharpParser.Object_creation_expressionContext):
        pass

    # Enter a parse tree produced by CSharpParser#identifier.
    def enterIdentifier(self, ctx: CSharpParser.IdentifierContext):
        pass

    # Exit a parse tree produced by CSharpParser#identifier.
    def exitIdentifier(self, ctx: CSharpParser.IdentifierContext):
        pass


del CSharpParser



"""
@file
@brief Shortcuts to sql
"""

from .database_helper import import_flatfile_into_database
from .database_main import Database
from .file_text_binary_columns import TextFileColumns
from .pandas_sql_helper import import_flatfile_into_database_pandas
from .sql_interface import InterfaceSQL, InterfaceSQLException



# -*- coding: utf-8 -*-
"""
@file
@brief Missing values and pandas.
"""
import pandas
import numpy
from .joins import df_crossjoin


def add_missing_indices(df, column, all_values, values=None, fillvalue=numpy.nan):
    """
    After aggregation, it usually happens that the series is sparse.
    This function adds rows for missing time.

    @param      df          dataframe to extend
    @param      column      column with time
    @param      all_values  all the values we want
    @param      values      columns which contain the values, the others are considered as the keys
    @return                 new dataframe

    .. exref::
        :title: Add missing values in one column.

        .. runpython::
            :showcode:

            import pandas
            from pyensae.mlhelper import add_missing_indices
            df = pandas.DataFrame([{"x": 3, "y": 4, "z": 1}, {"x": 5, "y": 6, "z": 2}])
            df2 = add_missing_indices(df, "x", [3, 4, 5, 6])
            print(df2)

        .. runpython::
            :showcode:

            import pandas
            from pyensae.mlhelper import add_missing_indices
            df = pandas.DataFrame([{"x": 3, "y": 4, "z": 1}, {"x": 5, "y": 6, "z": 2}])
            df2 = add_missing_indices(df, "x", values=["y"], all_values=[3, 4, 5, 6])
            print(df2)

    """
    if isinstance(values, str):
        values = [values]
    if values is None or len(values) == 0:
        keys = [_ for _ in df.columns if _ != column]
    else:
        keys = [_ for _ in df.columns if _ not in values and _ != column]
    if isinstance(all_values, list):
        dfti = pandas.DataFrame({column: all_values})
    elif isinstance(all_values, (pandas.Series, numpy.ndarray)):
        dfti = pandas.DataFrame({column: all_values})
    elif isinstance(all_values, pandas.DataFrame):
        dfti = all_values
        if dfti.shape[1] != 1:
            raise ValueError("all_values should have only one column")
        if dfti.columns[0] != column:
            raise ValueError(
                "all_values should have only one column with name '{0}'".format(column))
    else:
        raise TypeError(
            "Unexpected type for all_values '{0}'".format(type(all_values)))

    # Merge only happens on columns with the same type.
    cols = set(dfti.columns)
    for c in df.columns:
        if c in cols and dfti[c].dtype != df[c].dtype:
            dfti[c] = dfti[c].astype(df[c].dtype)

    if len(keys) == 0:
        dfj = df.merge(dfti, on=column, how="right")
    else:
        nkeys = keys + [column]
        only = df[nkeys].groupby(
            keys, as_index=False).count().drop(column, axis=1)
        dfti = df_crossjoin(only, dfti)
        dfj = df.merge(dfti, on=nkeys, how="right")
    return dfj.sort_values(column)



# -*- coding: utf-8 -*-
import sys
import os
from setuptools import find_packages, setup
from pyquicksetup import read_version, read_readme, default_cmdclass
from distutils.core import Command

#########
# settings
#########

project_var_name = "pyensae"
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = "HISTORY.rst"
requirements = None

KEYWORDS = [project_var_name, 'ENSAE', 'sqllite', 'database', 'teachings']
DESCRIPTION = """Helpers for teaching purposes (includes sqllite helpers)."""
CLASSIFIERS = [
    'Programming Language :: Python :: %d' % sys.version_info[0],
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]

#######
# data
#######


class SetupCommandBuildScript(Command):
    description = "Updates grammars."

    user_options = [
        ('g=', None, 'grammar file to recompile, R.g4 for example)')
    ]

    def initialize_options(self):
        self.g = None

    def finalize_options(self):
        pass

    def run(self):
        if self.g is None:
            raise RuntimeError(
                "Expecting a grammar file: python setup.py update_grammars R.g4")
        grammar = self.g
        from pyensae.languages import build_grammar
        if not os.path.exists(grammar):
            cdir = os.path.abspath(os.path.dirname(__file__))
            g2 = os.path.join(cdir, "src", "pyensae", "languages", grammar)
            if not os.path.exists(g2):
                raise FileNotFoundError("{0}\n{1}".format(grammar, g2))
            grammar = g2
        build_grammar(grammar, fLOG=print)
        r = True


packages = find_packages('src', exclude='src')
package_dir = {k: "src/" + k.replace(".", "/") for k in packages}
package_data = {project_var_name + ".subproject": ["*.tohelp"],
                project_var_name + ".languages": ["*.g4", "*.tokens"], }
command = default_cmdclass().copy()
command['update_grammars'] = SetupCommandBuildScript


setup(
    name=project_var_name,
    version=read_version(__file__, project_var_name, subfolder='src'),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url="http://www.xavierdupre.fr/app/pyensae/helpsphinx/index.html",
    download_url="https://github.com/sdpython/pyensae/",
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=command,
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=["pyquicksetup"],
    install_requires=[
        "pyquickhelper>=1.8",
        "numpy>=1.18", "pandas>=1.0", "matplotlib>=3.0"
    ],
    extras_require={
        'graphhelper': ['blockdiag', 'cartopy'],
        'datasource': ['dbread', 'geopandas', 'shapely', 'pyshp', 'pylzma'],
        'finance': ['pandas-datareader', 'yahoo_historical>=0.4'],
        'graphhelper': ['matplotlib>=3.0'],
        'languages': ['antlr4-python3-runtime>=4.10'],
        'notebookhelper': ['folium', 'qgrid'],
        'all': [
            "pyquickhelper>=1.8", "numpy>=1.18", "pandas>=1.0", "matplotlib>=3.0",
            'blockdiag', 'cartopy',
            'dbread', 'geopandas', 'shapely', 'pyshp', 'pylzma',
            'pandas-datareader', 'yahoo_historical>=0.4',
            'antlr4-python3-runtime>=4.8',
            'folium', 'qgrid', 'easydev',
        ],
    },
    entry_points={
        'console_scripts': [
            'file_head = pyensae.cli.head_cli:file_head_cli',
            'file_tail = pyensae.cli.tail_cli:file_tail_cli',
        ]}
)



"""
@file
@brief Geographic datasets.
"""
import math
import os
import pandas
from pyquickhelper.filehelper import un7zip_files
from .http_retrieve import download_data


def lambert932WGPS(lambertE, lambertN):
    """
    Converts Lambert coordinates into longitude, latitude.
    Does not use :epkg:`pyproj`.

    @param      lambertE        lambertE
    @param      lambertN        lambertN
    @return                     longitude, latitude
    """

    class constantes:
        GRS80E = 0.081819191042816
        LONG_0 = 3
        XS = 700000
        YS = 12655612.0499
        n = 0.7256077650532670
        C = 11754255.4261

    delX = lambertE - constantes.XS
    delY = lambertN - constantes.YS
    gamma = math.atan(-delX / delY)
    R = math.sqrt(delX * delX + delY * delY)
    latiso = math.log(constantes.C / R) / constantes.n
    sinPhiit0 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * math.sin(1)))
    sinPhiit1 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit0))
    sinPhiit2 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit1))
    sinPhiit3 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit2))
    sinPhiit4 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit3))
    sinPhiit5 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit4))
    sinPhiit6 = math.tanh(latiso + constantes.GRS80E *
                          math.atanh(constantes.GRS80E * sinPhiit5))

    longRad = math.asin(sinPhiit6)
    latRad = gamma / constantes.n + constantes.LONG_0 / 180 * math.pi

    longitude = latRad / math.pi * 180
    latitude = longRad / math.pi * 180

    return longitude, latitude


def load_french_departements(cache=None):
    """
    Returns the definition of French departments.

    @param      cache       cache folder
    @return                 French departments as a :epkg:`GeoDataFrame`

    .. exref::
        :title: Loads French departments polygons

        Simple example to retrieve French departements.

        .. runpython::
            :showcode:

            from pyensae.datasource import load_french_departements
            df = load_french_departements()
            print(df.head(2).T)
    """
    if cache is None:
        cache = '.'
    # delayed import
    import shapefile
    import geopandas
    from shapely.geometry import Polygon
    from shapely.ops import unary_union

    name = "GEOFLA_2-1_DEPARTEMENT_SHP_LAMB93_FXX_2015-12-01.7z"
    try:
        download_data(name, whereTo=cache,
                      website="https://wxs-telechargement.ign.fr/oikr5jryiph0iwhw36053ptm/telechargement/inspire/"
                      "GEOFLA_THEME-DEPARTEMENTS_2015_2$GEOFLA_2-1_DEPARTEMENT_SHP_LAMB93_FXX_2015-12-01/file/")
    except Exception as e:
        # au cas le site n'est pas accessible
        download_data(name, website="xd", whereTo=cache)

    full_name = os.path.join(cache, name)
    un7zip_files(full_name, where_to=os.path.join(cache, "shapefiles"))
    departements = os.path.join(cache, 'shapefiles/GEOFLA_2-1_DEPARTEMENT_SHP_LAMB93_FXX_2015-12-01/GEOFLA/1_DONNEES_LIVRAISON_2015/'
                                'GEOFLA_2-1_SHP_LAMB93_FR-ED152/DEPARTEMENT/DEPARTEMENT.shp')

    r = shapefile.Reader(departements)
    shapes = r.shapes()
    records = r.records()

    polys = []
    datas = []
    for i, (record, shape) in enumerate(zip(records, shapes)):
        # coordinates in Lambert 93
        geo_points = [lambert932WGPS(x, y) for x, y in shape.points]
        rec = {}
        for k in ['CODE_DEPT', 'CODE_REG', 'CODE_CHF', 'ID_GEOFLA', 'NOM_CHF',
                  'NOM_DEPT', 'NOM_REG', 'X_CENTROID', 'X_CHF_LIEU',
                  'Y_CENTROID', 'Y_CHF_LIEU']:
            rec[k] = getattr(record, k, None)
        if len(shape.parts) == 1:
            # only one polygon
            poly = Polygon(geo_points)
        else:
            # merge them
            ind = list(shape.parts) + [len(shape.points)]
            pols = [Polygon(geo_points[ind[i]:ind[i + 1]])
                    for i in range(0, len(shape.parts))]
            try:
                poly = unary_union(pols)
            except Exception as e:
                poly = Polygon(geo_points)
        polys.append(poly)
        datas.append(rec)

    geom = geopandas.GeoDataFrame(geometry=polys)
    data = pandas.DataFrame(datas)
    return pandas.concat([geom, data], axis=1)



# -*- coding: utf-8 -*-
"""
@file
@brief Magic command to help within notebooks
"""
from IPython.core.magic import magics_class, line_magic
try:
    import qgrid
except ImportError:
    # We ignore that error since it is not
    # part of the mandatory requirements.
    qgrid = None
from jyquickhelper import add_notebook_menu
from pyquickhelper.ipythonhelper import MagicCommandParser, MagicClassWithHelpers


@magics_class
class MagicNotebook(MagicClassWithHelpers):

    """
    Defines magic commands to help with notebooks

    .. versionadded:: 1.1
    """

    @staticmethod
    def nb_menu_parser():
        """
        defines the way to parse the magic command ``%nb_menu``
        """
        parser = MagicCommandParser(
            description='display a menu in the notebook based on javascript', prog="nb_menu")
        parser.add_argument(
            '-t',
            '--title',
            type=str,
            default="Plan",
            help='title before the menu')
        parser.add_argument(
            '-i',
            '--menu-id',
            type=str,
            default="my_menu_id",
            help='menu id used to locate the corresponding HTML tag div')
        parser.add_argument(
            '-f',
            '--format',
            type=str,
            default="html",
            help='format to consider, html or rst')
        parser.add_argument(
            '-l1',
            '--level1',
            type=int,
            default=2,
            help='first level to consider in the menu')
        parser.add_argument(
            '-l2',
            '--level2',
            type=int,
            default=4,
            help='last level to consider in the menu')
        parser.add_argument(
            '-r',
            '--raw',
            type=bool,
            default=False,
            help='if true, displays the javascript, otherwise the html')
        return parser

    @line_magic
    def nb_menu(self, line):
        """
        defines ``%nb_menu``
        which displays a menu

        .. nbref::
            :title: nb_menu

            The magic command ``%nb_menu`` displays a menu in a notebook.
            It parses a notebook and displays a menu gathering all sections
            within that notebook. The code is the following::

                from pyquickhelper.ipythonhelper import add_notebook_menu
                js = add_notebook_menu(menu_id=<menu_id>, header=<title>, format=<format>,
                                    first_level=<level1>, last_level=<level2>,
                                    raw=<raw>)
        """
        parser = self.get_parser(MagicNotebook.nb_menu_parser, "nb_menu")
        args = self.get_args(line, parser)

        if args is not None:
            js = add_notebook_menu(menu_id=args.menu_id, header=args.title, format=args.format,
                                   first_level=args.level1, last_level=args.level2, raw=args.raw)
            return js

    @staticmethod
    def jsdf_parser():
        """
        defines the way to parse the magic command ``%jsdf``
        """
        parser = MagicCommandParser(
            description='display a pandas DataFrame based on module qgrid', prog="jsdf")
        parser.add_argument(
            'df',
            help='dataframe to display')
        parser.add_argument(
            '--defaultColumnWidth',
            type=int,
            default=80,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--enableColumnReorder',
            type=bool,
            default=True,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--multiColumnSort',
            type=bool,
            default=False,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--rowHeight',
            type=int,
            default=25,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--showHeaderRow',
            type=bool,
            default=False,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--forceFitColumns',
            type=bool,
            default=False,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--autoHeight',
            type=bool,
            default=False,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--enableCellNavigation',
            type=bool,
            default=True,
            help='see https://github.com/mleibman/SlickGrid/wiki/Grid-Options')
        parser.add_argument(
            '--precision',
            type=int,
            default=4,
            help='see function qgrid.set_defaults')
        parser.add_argument(
            '--editable',
            type=bool,
            default=True,
            help='make the output editable')
        return parser

    @line_magic
    def jsdf(self, line):
        """
        defines ``%jsdf``
        which displays a pandas dataframe into a notebook using qgrid (javascript)

        .. nbref::
            :title: jsdf

            The magic command ``%jsdf`` displays a dataframe using
            `qgrid <https://qgrid.readthedocs.io/en/latest/>`_ module.
            The code is the following::

                import qgrid
                if firt_call:
                    qgrid.set_defaults(precision=<precision>)
                    self.first_jsdf_call = False

                df = args.df
                grid_options = dict(defaultColumnWidth=<defaultColumnWidth>,
                                    enableColumnReorder=<enableColumnReorder>,
                                    multiColumnSort=<multiColumnSort>,
                                    rowHeight=<rowHeight>,
                                    showHeaderRow=<showHeaderRow>,
                                    forceFitColumns=<forceFitColumns>,
                                    autoHeight=<autoHeight>,
                                    enableCellNavigation=<enableCellNavigation>)
                qgrid.show_grid(df, grid_options=grid_options)
        """
        parser = self.get_parser(MagicNotebook.jsdf_parser, "jsdf")
        args = self.get_args(line, parser)

        if qgrid is None:
            return "qgrid is not installed."

        if not hasattr(self, "first_jsdf_call") or self.first_jsdf_call:
            if args is not None:
                qgrid.set_defaults(precision=args.precision)
            else:
                qgrid.set_defaults()
            self.first_jsdf_call = False

        if args is not None:
            df = args.df
            grid_options = dict(defaultColumnWidth=args.defaultColumnWidth,
                                enableColumnReorder=args.enableColumnReorder,
                                multiColumnSort=args.multiColumnSort,
                                rowHeight=args.rowHeight,
                                showHeaderRow=args.showHeaderRow,
                                forceFitColumns=args.forceFitColumns,
                                autoHeight=args.autoHeight,
                                enableCellNavigation=args.enableCellNavigation,
                                editable=args.editable)
            res = qgrid.show_grid(df, grid_options=grid_options)
            return "" if res is None else res


def register_notebook_magics(ip=None):
    """
    register magics function, can be called from a notebook

    @param      ip      from ``get_ipython()``
    """
    if ip is None:
        from IPython import get_ipython
        ip = get_ipython()
    ip.register_magics(MagicNotebook)



"""
@file
@brief Various functions to display information about files
"""

import datetime


def format_file_size(size):
    """
    Formats the file size as string.

    @param      size        numeric value
    @return                 string (something + unit)
    """
    if size >= 2 ** 30:
        size = size / 2 ** 30
        return "%1.2f Gb" % size
    elif size >= 2 ** 20:
        size = size / 2 ** 20
        return "%1.2f Mb" % size
    elif size >= 2 ** 10:
        size = size / 2 ** 10
        return "%1.2f Kb" % size
    else:
        return "%d" % size


def format_file_mtime(timestamp):
    """
    Returns a :epkg:`datetime` for a file.

    @param      timestamp       modified date for example
    @return                     datetime
    """
    return datetime.datetime.fromtimestamp(timestamp)



"""
@file

@brief @see cl Database
"""

import re
import copy
import random
import sqlite3 as SQLite
import datetime
import decimal
import numpy


from .file_text_binary import TextFile
from .database_exception import ExceptionSQL


class NoHeaderException(Exception):

    """
    just to be meant to be caucht later by a unit test
    """
    pass


class DatabaseCore2:

    """
    Complementary methods for class @see cl Database.
    """

    _split_expr = "\\r?\\t"

    def _check_connection(self):
        """
        Checks the SQL connection.
        """
        if "_connection" not in self.__dict__:
            message = "use connect method before doing operation on this database"
            raise Exception(message)

    def _check_values(self, values):
        """
        When values are inserted or updated, this method doubles ``"'"``
        it does not allow str values, only str.

        @param      values      dictionary
        @return                 dictionary
        """
        mod = []
        for k, v in values.items():
            if isinstance(v, str) and "'" in v:
                mod.append(k)
        if len(mod) == 0:
            return values
        else:
            values = copy.copy(values)
            for k in mod:
                values[k] = values[k].replace("'", "''")
            return values

    def summary(self, light=False):
        """
        Returns the list of tables, their columns, and their length.

        @param      light   light version, no count, no first lines
        @return             a dictionary where the keys are (t,i), t is a table name, i is in ["columns", "size", "first_lines"],
                            a str message
        """
        tables = self.get_table_list()
        indexes = self.get_index_list()
        res = {}
        lines = []

        for t in tables:
            col = self.get_table_columns_list(t)
            if not light:
                size = self.get_table_nb_lines(t)
                first = self.get_table_nfirst_lines(t)
            else:
                size = -1
                first = []

            res[t, "columns"] = col
            res[t, "size"] = size
            res[t, "first_lines"] = first

            lines.append(t + "\t" + str(size) + " records")
            lines.append("    columns")
            for c in col:
                lines.append("        " + str(c))

            if len(first) > 0:
                lines.append("    first_lines")
                for lf in first:
                    fo = []
                    if lf is None:
                        lines.append("        None")
                    else:
                        for x in lf:
                            if not isinstance(x, str):
                                fo.append(str(x))
                            else:
                                fo.append(x)
                        lines.append("        " + "\t".join(fo))

        if len(indexes) > 0:
            lines.append("\n")
            lines.append("indexes")
            for tu in indexes:
                if isinstance(tu, (tuple, list)):
                    lines.append("    " + "\t".join([str(x) for x in tu]))
                else:
                    lines.append("    " + tu)

        attached = self.get_attached_database_list()
        if len(attached) > 0:
            lines.append("\n")
            lines.append("attached databases")
            for a in attached:
                if a == "main":
                    continue
                lines.append("    " + "\t" + a)
                continue
                # ~ rrr = self.execute(
                #       ~ "SELECT name FROM %s.sqlite_master ORDER BY name;" %
                #       ~ (a,))
                # ~ for b in rrr:
                #       ~ lines.append("       " + "\t" + b[0])

        return res, "\n".join(lines)

    def _guess_columns(
            self, file, format, columns_name=None, filter_case=None, header=True, encoding="utf-8"):
        """
        Guesses the columns types from a file (the method assumes there is a header),
        The types are chosen in that order: int, float, str.
        It keeps the most frequent one with if there is not too many errors.
        The separator must be tabs (``\\t``).

        @param      file            file name
        @param      format          format (only tsv)
        @param      columns_name    if None, the first line contains the columns, otherwise it is the columns name
        @param      filter_case     process every case information (used to replace space for example)
        @param      header          by default, the function is expected a header
        @param      encoding        encoding
        @return                     columns, changes
        """
        f = TextFile(file, fLOG=self.LOG, encoding=encoding)
        f.open()

        if header:
            _aa, _bb, _cc, _dd = f.guess_columns(fields=columns_name)
            reg_exp = re.compile(DatabaseCore2._split_expr.replace(
                "\\t", _cc.replace("|", "[|]")))
        else:
            # tabulation by default
            reg_exp = re.compile(DatabaseCore2._split_expr)
            f.close()
            raise NoHeaderException("a header is expected for that function")

        self.LOG("  [_guess_columns] sep={0}".format([_cc]))

        lines = []
        for line in f:
            if len(lines) > 1000:
                break
            if len(lines) > 900 and random.randint(0, 10) > 0:
                continue
            lines.append(reg_exp.split(
                line.strip(" \r\n").strip('\ufeff')))
        f.close()

        if len(lines) <= 1:
            raise Exception("file %s is empty" % file)

        exp = re.compile("\\W+")
        columns = {}
        done = {}
        count = {}
        changes = {}

        for i in range(0, len(lines[0])):
            if lines[0][i] in [
                    '\ufeffID', '\ufeffid', '\ufeffqid', '\ufeffQID']:
                lines[0][i] = "qid"

            if columns_name is None:
                name = lines[0][i].replace(":", "_")
                origin = lines[0][i]
            else:
                name = columns_name[i].replace(":", "_")
                origin = columns_name[i]

            name = name.replace("-", "_").replace(" ", "_")

            spl = exp.split(name)
            if len(spl) > 1:
                name = "".join(spl)
            if name[0] in "0123456789":
                name = "_" + name

            if name in count:
                count[name] += 1
                name += str(count[name])
            else:
                count[name] = 1

            #lines [0][i] = name
            columns[i] = (name, int)
            done[i] = False

            if origin != name:
                changes[origin] = name

        self.LOG("  [_guess_columns] columns_name={0}".format(columns_name))

        length = {}
        nbline = 0
        count_types = {}

        for line_ in lines[1:]:
            if filter_case is None:
                line = line_
            else:
                line = [filter_case(s) for s in line_]
            nbline += 1
            if line == [] or line == ['']:
                continue

            for i in range(0, len(line)):

                if i >= len(done):
                    # it is probably a wrong line
                    continue

                vl = length.get(i, 0)
                if len(line[i]) > vl:
                    length[i] = len(line[i])

                try:
                    if done[i]:
                        continue
                except KeyError as e:
                    str_columns = ""
                    for k, v in columns.items():
                        str_columns += "       " + \
                            str(k) + "\t" + str(v) + "\n"
                    mes = "KeyError:" + str(e) + "\n" + str(done) + "\n" + str_columns + "\nnb line " + str(
                        nbline) + " columns: " + str(len(line)) + "\n" + str(line)
                    raise RuntimeError(  # pylint: disable=W0707
                        "problem\n" +
                        mes +
                        "\n\ncount_types:\n  " +
                        "\n  ".join(
                            "{0}:{1}".format(
                                k,
                                v) for k,
                            v in sorted(
                                count_types.items())))

                if line[i] is None or len(line[i]) == 0:
                    continue

                try:
                    x = int(line[i])
                    if abs(x) >= 2147483647:
                        raise ValueError("too big int")

                    if i not in count_types:
                        count_types[i] = {int: 1}
                    else:
                        count_types[i][int] = count_types[i].get(int, 0) + 1

                except ValueError:
                    try:
                        x = float(line[i])

                        if i not in count_types:
                            count_types[i] = {float: 1}
                        else:
                            count_types[i][float] = count_types[
                                i].get(float, 0) + 1

                        if columns[i][1] != float:
                            columns[i] = (columns[i][0], float)

                    except ValueError:
                        columns[i] = (
                            columns[i][0], (str, max(
                                1, len(
                                    line[i])) * 2))

                        if i not in count_types:
                            count_types[i] = {str: 1}
                        else:
                            count_types[i][str] = count_types[
                                i].get(str, 0) + 1

        self.LOG("   guess with ", len(lines), "lines")
        self.LOG("   count_types ", count_types)
        for i in range(0, len(columns)):

            # if i is not in count_types, it means the first rows do now
            # contain values for these columns (only null values)
            t = count_types.get(i, {str: 1})
            nb = sum(t.values())

            th = 0.0 if nb < 50 else (
                0.01 if nb < 100 else 0.02)  # we authorize 2% of wrong types

            n = t.get(int, 0)
            if n * 1.0 / nb >= 1 - th:
                ty = int
            else:
                n += t.get(float, 0)
                if n * 1.0 / nb >= 1 - th:
                    ty = float
                else:
                    ty = str

            columns[i] = (columns[i][0], ty)

        self.LOG("   columns ", columns)

        # if not done, choose str by default
        for c in columns:
            v = columns[c]
            if v[1] == str:
                columns[c] = (v[0], (str, max(1, length.get(c, 4)) * 2))

        for c, v in columns.items():
            t = v[1]
            if isinstance(t, tuple) and t[0] == str and t[1] == 0:
                raise Exception(
                    "the length is null for column %s - %s" %
                    (c, str(v)))

        self.LOG("   guess", columns)
        return columns, changes

    def _process_text_line(self, line, columns, format, lower_case, num_line,
                           fill_missing=0, filter_case=None,
                           strict_separator=False):
        """
        Processes a text line.

        @param      line                text line to process (or a list if it already splitted)
        @param      columns             columns definition @see me _append_table
        @param      format              only tsv for the moment
        @param      lower_case          put every str object in lower_case
        @param      num_line            line number
        @param      fill_missing        fill the missing values by a default value, at least not more than fill_missing values
        @param      filter_case         process every case information (used to replace space for example)
        @param      strict_separator    strict number of columns, it assumes there is no separator in the content of every column
        @return                         a dictionary
        """
        if not isinstance(line, list) and not isinstance(
                line, tuple) and not isinstance(line, numpy.ndarray):
            if format != "tsv":
                raise Exception("unable to process format " + format)
            line = line.strip("\r\n ").replace("\n", " ")
            line = DatabaseCore2._split_expr.split(line)

        if filter_case is not None:
            line = [filter_case(s) for s in line]

        try:
            if fill_missing > 0:
                m = max(columns.keys())
                if m >= len(line):
                    line = copy.copy(line)
                    add = 0
                    while m >= len(line) and add < fill_missing:
                        a, b = columns[len(line)]
                        if b is int:
                            line.append("0")
                        elif b is float:
                            line.append("0.0")
                        elif b is decimal.Decimal:
                            line.append("0")
                        elif b is str:
                            line.append("")
                        else:
                            line.append("")
                        add += 1

            res = {}
            for c, v in columns.items():
                if "AUTOFILL" in v:
                    res[v[0]] = "NULL"
                elif "AUTOINCREMENT" in v:
                    continue
                else:
                    if c >= len(line):
                        self.LOG(
                            "(a)line number ",
                            num_line,
                            "*unable to process a line columns ",
                            c,
                            "#",
                            line,
                            " columns ",
                            columns)
                        return None

                    val = line[c]
                    if len(v) > 2 and v[2].lower() not in [
                            "primarykey", "autofill"]:
                        val = v[2](val)

                    try:
                        if isinstance(v[1], tuple):
                            val = v[1][0](val)
                        elif v[1] is datetime.datetime:
                            if isinstance(val, datetime.datetime):
                                pass
                            elif isinstance(val, str):
                                val = datetime.datetime.parse(val)
                            else:
                                raise TypeError(
                                    "unable to convert %s into datetime" % str(
                                        type(val)))
                        else:
                            val = v[1](val)
                    except ValueError:  # as e :
                        self.LOG(
                            "(b)line number ",
                            num_line,
                            "**unable to process a line columns ",
                            c,
                            "#",
                            v[0],
                            " type ",
                            v[1],
                            " value ",
                            repr(
                                line[c]))
                        return None

                    if isinstance(val, str):
                        val = val.replace("'", "''")
                        if lower_case:
                            val = val.lower()
                    res[v[0]] = val

            return res
        except Exception:
            self.LOG("(c)line number", num_line,
                     "***unable to process a line columns:", line)
            return None

    def _get_insert_request(self, dico,
                            table,
                            exe=False,
                            primarykey=None,
                            cursor=None,
                            skip_exception=False):
        """
        Builds an ``INSERT SQL`` request from a dictionary.

        @param      dico            dictionary
        @param      table           table name
        @param      exe             if True, execute the request, if False, do nothing except returning the request
        @param      primarykey      primary key column, if it exist
        @param      cursor          if None, creates a new one, otherwise use it
        @param      skip_exception  if True, log exception instead of raising one
        @return                     str """
        keys = []
        values = []
        for k, v in dico.items():
            keys.append(k)
            if k != primarykey and isinstance(v, str):
                v = "'" + str(v).replace("'", "''") + "'"
                values.append(v)
            elif isinstance(v, datetime.datetime):
                values.append("'" + str(v) + "'")
            else:
                values.append(str(v))
        keys = ",".join(keys)
        values = ",".join(values)
        sql = "INSERT INTO %s (%s) VALUES (%s);" % (table, keys, values)
        if exe:
            try:
                if cursor is not None:
                    cursor.execute(sql)
                else:
                    self._connection.execute(sql)
            except SQLite.OperationalError as e:
                if skip_exception:
                    self.LOG(
                        "OperationalError: unable to execute a query", e, sql)
                else:
                    raise ExceptionSQL(  # pylint: disable=W0707
                        "OperationalError: unable to execute a query", e, sql)
            except SQLite.IntegrityError as e:
                if skip_exception:
                    self.LOG("IntegrityError: unable to execute a query", e, sql)
                else:
                    raise ExceptionSQL(  # pylint: disable=W0707
                        "IntegrityError: unable to execute a query", e, sql)
        return sql

    def get_python_code(self, varname="db"):
        """
        Returns the python code associated to this database.

        @param      varname     name of the variable
        @return                 2-uple: simp, scode (import part, code part)
        """
        simp = ["from pyensae import Database"]

        code = ["tblname = r'%s'" % self.get_file()]

        more = []
        if self._engine != "SQLite":
            code.append("engine = '%s'" % self._engine)
        if self._user is not None:
            code.append("user = '%s'" % self._user)
        if self._password is not None:
            code.append("password = '%s'" % self._password)
        if self._host is not None:
            code.append("host = '%s'" % self._host)
        more = ", ".join(more)

        code.append(
            "%s      = Database (dbfile = tblname%s)" %
            (varname, more))
        code.append("%s.connect ()" % varname)
        att = self.get_attached_database_list(file=True)
        for alias, file in att:
            if len(file) == 0:
                continue
            code.append(
                "%s.attach_database ('%s','%s')" %
                (varname, file, alias))

        return "\n".join(simp) + "\n", "\n".join(code) + "\n"



# -*- coding: utf-8 -*-
"""
@file
@brief Module *folium* does not have any output to a notebook, addresses that issue. The module
does not explicitely import *folium*.
"""

from IPython.display import HTML


def folium_html_map(mapf, width=None, height=None, asobj=True):
    """
    Embeds the HTML source of the map directly into the IPython notebook.

    @param      mapf    folium map
    @param      width   width
    @param      height  height
    @param      asobj   return an object which implements ``_repr_html_``
    @return             HTML (IPython)

    This method will not work if the map depends on any files (json data). Also this uses
    the HTML5 srcdoc attribute, which may not be supported in all browsers.

    Source: `folium_base.py <https://gist.github.com/psychemedia/f7385255f89137c503b5>`_

    .. exref::
        :title: Display an inline map with folium in a notebook

        ::

            import folium
            map_osm = folium.Map(location=[48.85, 2.34])
            from pyensae.notebook_helper import folium_html_map
            map_osm.polygon_marker(location=[48.824338, 2.302641], popup='ENSAE',
                                fill_color='#132b5e', num_sides=3, radius=10)
            folium_html_map(map_osm)

        With folium version 0.2, this becomes easier:

        ::

            import folium
            map_osm = folium.Map(location=[48.85, 2.34])
            from pyensae.notebook_helper import folium_html_map
            map_osm.polygon_marker(location=[48.824338, 2.302641], popup='ENSAE',
                                fill_color='#132b5e', num_sides=3, radius=10)
            map_osm

    .. versionchanged:: 1.1
        Add parameters *width* and *height* to change the size of the map within a notebook.
        Hopefully, they will be added in folium.
    """
    res = mapf._repr_html_()
    if width or height:
        look = '<div style="width:100%;">'
        if not res.startswith(look):
            raise ValueError(
                "Folium has changed its HTML form, it used to start with: '{0}'.\n{1}".format(look, res))
        size = ""
        if width:
            size += "width:" + width + ";"
        if height:
            size += "height:" + height + ";"
        newlook = '<div style="{size}">'.format(size=size)
        res = newlook + res[len(look):]
    if asobj:
        class CustomFoliumMap:

            def __init__(self, res, map):
                self.res = res
                self.map = map

            def _repr_html_(self):
                return self.res

        return CustomFoliumMap(res, map)
    else:
        return res


def folium_embed_map(mapf, path="map.html", width="100%", height="510px"):
    """
    @param      mapf    folium map
    @param      path    where to store the temporary map
    @return             HTML (IPython)

    Embeds a linked iframe to the map into the IPython notebook.

    Note: this method will not capture the source of the map into the notebook.
    This method should work for all maps (as long as they use relative urls).

    Source: `folium_base.py <https://gist.github.com/psychemedia/f7385255f89137c503b5>`_
    """
    mapf.save(path)
    return HTML('<iframe src="files/{path}" style="width: {width}; height: {height}; border: none">' +
                '</iframe>'.format(path=path, width=width, height=height))



# -*- coding: utf-8 -*-
"""
@file
@brief Magic command to handle files
"""
import sys
import os
import pandas

from IPython.core.magic import magics_class, line_magic, cell_magic
from IPython.core.display import HTML

from pyquickhelper.filehelper.synchelper import explore_folder_iterfile, explore_folder_iterfile_repo
from pyquickhelper.loghelper import run_cmd
from pyquickhelper.ipythonhelper import MagicCommandParser, MagicClassWithHelpers
from pyquickhelper.helpgen import docstring2html
from .format_helper import format_file_size, format_file_mtime
from .content_helper import file_head, file_tail, enumerate_grep, file_encoding


@magics_class
class MagicFile(MagicClassWithHelpers):

    """
    Defines magic commands to help with files

    .. faqref::
        :title: Magic command not found

        Magic commands are automatically added when importing
        the module::

            import pyensae

        This instruction does not raise any exception.
        To understand what went wrong, the following instruction must be run::

            %load_ext pyensae

        Usually, one necessary module is missing.

    .. versionadded:: 1.1
    """

    @staticmethod
    def head_parser():
        """
        defines the way to parse the magic command ``%head``
        """
        parser = MagicCommandParser(prog="head",
                                    description='display the first lines of a text file')
        parser.add_argument('f', type=str, help='filename')
        parser.add_argument(
            '-n',
            '--n',
            type=int,
            default=10,
            help='number of lines to display')
        parser.add_argument(
            '-r',
            '--raw',
            default=False,
            action='store_true',
            help='display raw text instead of HTML')
        parser.add_argument(
            '-e',
            '--encoding',
            default="utf8",
            help='file encoding')
        parser.add_argument(
            '-s',
            '--errors',
            default="",
            help='What about errors: "", strict, replace, surrogateescape, '
                 'xmlcharrefreplace, backslashreplace, namereplace')
        return parser

    @line_magic
    def head(self, line):
        """
        Defines ``%head``
        which displays the first lines of a file.

        .. nbref::
            :tag: file
            :title: head

            The magic command ``%head`` is equivalent to::

                from pyensae.file_helper import file_head
                file_head(<filename>, <n>, <encoding>)
        """
        parser = self.get_parser(MagicFile.head_parser, "head")
        args = self.get_args(line, parser)

        if args is not None:
            errors = args.errors
            if errors is not None and len(errors) < 2:
                errors = None
            encoding = args.encoding
            if encoding is not None and len(encoding) < 2:
                encoding = None
            rows = file_head(args.f, args.n, encoding, errors=errors)
            if args.raw:
                return "".join(rows)
            else:
                return HTML("<pre>\n{0}\n</pre>".format("".join(rows)))
        return None

    @staticmethod
    def encoding_parser():
        """
        Defines the way to parse the magic command ``%encoding``.
        """
        parser = MagicCommandParser(prog="encoding",
                                    description='guess the encoding of a file')
        parser.add_argument('f', type=str, help='filename')
        parser.add_argument(
            '-n',
            '--n',
            type=int,
            default=2**20,
            help='maximum number of lines to use to guess the encoding')
        return parser

    @line_magic
    def encoding(self, line):
        """
        Defines ``%encoding``
        which guesses the encoding.

        .. nbref::
            :tag: file
            :title: encoding

            The magic command ``%encoding`` is equivalent to::

                from pyensae.file_helper import file_head
                file_head(<filename>, <n>, <encoding>)
        """
        parser = self.get_parser(MagicFile.encoding_parser, "encoding")
        args = self.get_args(line, parser)

        if args is not None:
            res = file_encoding(args.f, args.n)
            return str(res)
        return None

    @staticmethod
    def grep_parser():
        """
        defines the way to parse the magic command ``%grep``
        """
        parser = MagicCommandParser(prog="grep",
                                    description='display the first lines of a text file')
        parser.add_argument('f', type=str, help='filename')
        parser.add_argument('regex', type=str, help='regular expression')
        parser.add_argument(
            '-n',
            '--n',
            type=int,
            default=-1,
            help='number of lines to display, -1 for all')
        parser.add_argument(
            '-r',
            '--raw',
            default=False,
            action='store_true',
            help='display raw text instead of HTML')
        parser.add_argument(
            '-e',
            '--encoding',
            default="utf8",
            help='file encoding')
        parser.add_argument(
            '-s',
            '--errors',
            default="",
            help='What about errors: "", strict, replace, surrogateescape, '
                 'xmlcharrefreplace, backslashreplace, namereplace')
        return parser

    @line_magic
    def grep(self, line):
        """
        defines ``%grep``
        which displays the first lines of a file

        .. nbref::
            :tag: file
            :title: grep

            The magic command ``%grep`` is equivalent to::

                from pyensae.file_helper import enumerate_grep
                list(enumerate_grep(<filename>, <regex>, <encoding>))
        """
        parser = self.get_parser(MagicFile.grep_parser, "grep")
        args = self.get_args(line, parser)

        if args is not None:
            errors = args.errors
            if errors is not None and len(errors) < 2:
                errors = None
            encoding = args.encoding
            if encoding is not None and len(encoding) < 2:
                encoding = None
            iter = enumerate_grep(args.f, args.regex,
                                  encoding, errors=errors)
            if args.n != -1:
                rows = []
                for r in iter:
                    if len(rows) >= args.n:
                        break
                    rows.append(r)
            else:
                rows = list(iter)

            if args.raw:
                return "".join(rows)
            else:
                return HTML("<pre>\n{0}\n</pre>".format("".join(rows)))
        return None

    @staticmethod
    def tail_parser():
        """
        defines the way to parse the magic command ``%tail``
        """
        parser = MagicCommandParser(prog="tail",
                                    description='display the last lines of a text file')
        parser.add_argument('f', type=str, help='filename')
        parser.add_argument(
            '-n',
            '--n',
            type=int,
            default=10,
            help='number of lines to display')
        parser.add_argument(
            '-r',
            '--raw',
            default=False,
            action='store_true',
            help='display raw text instead of HTML')
        parser.add_argument(
            '-e',
            '--encoding',
            default="utf8",
            help='file encoding')
        parser.add_argument(
            '-s',
            '--errors',
            default="",
            help='What about errors: "", strict, replace, surrogateescape, '
                 'xmlcharrefreplace, backslashreplace, namereplace')
        return parser

    @line_magic
    def tail(self, line):
        """
        defines ``%tail``
        which displays the last lines of a file

        .. nbref::
            :tag: file
            :title: tail

            The magic command ``%tail`` is equivalent to::

                from pyensae.file_helper import file_tail
                file_tail(<filename>, <n>, <encoding>)
        """
        parser = self.get_parser(MagicFile.tail_parser, "tail")
        args = self.get_args(line, parser)

        if args is not None:
            errors = args.errors
            if errors is not None and len(errors) < 2:
                errors = None
            encoding = args.encoding
            if encoding is not None and len(encoding) < 2:
                encoding = None
            rows = file_tail(args.f, args.n, encoding, errors=errors)
            if args.raw:
                return "".join(rows)
            else:
                return HTML("<pre>\n{0}\n</pre>".format("".join(rows)))
        return None

    @staticmethod
    def lsr_parser():
        """
        defines the way to parse the magic command ``%lsr``
        """
        parser = MagicCommandParser(prog="lsr",
                                    description='display the content of a folder as a dataframe')
        parser.add_argument(
            'path',
            type=str,
            nargs="?",
            help='path',
            default=".")
        parser.add_argument(
            '-f',
            '--filter',
            type=str,
            default=".*",
            help='filter, same syntax as a regular expression')
        return parser

    @line_magic
    def lsr(self, line):
        """
        Defines ``%lsr`` which returns the content of a folder,
        the method stops after around 10000 files
        --> you should precise the filter.

        .. nbref::
            :tag: file
            :title: lsr

            The magic command ``%lsr`` is almost equivalent to::

                from pyquickhelper.file_helper import explore_folder_iterfile
                res = explore_folder_iterfile(<filename>, <pattern>)
                for f in res:
                    print(f)
        """
        parser = self.get_parser(MagicFile.lsr_parser, "lsr")
        args = self.get_args(line, parser)

        if args is not None:
            if args.path is None or len(args.path) == 0:
                filename = "."
            else:
                filename = args.path
            pattern = args.filter

            if "*" in filename:
                pattern = filename
                filename = "."

            iter = explore_folder_iterfile(filename, pattern)
            rows = []
            for r in iter:
                d = os.path.isfile(r)
                if d:
                    st = os.stat(r)
                    r = {"name": r,
                         "size": format_file_size(st.st_size),
                         "last_modified": format_file_mtime(st.st_mtime),
                         "directory": False}
                else:
                    r = {"name": r, "directory": True}
                rows.append(r)
            return pandas.DataFrame(rows)
        return None

    @staticmethod
    def PYTHON_parser():
        """
        Defines the way to parse the magic command ``%%PYTHON``.
        """
        parser = MagicCommandParser(prog="PYTHON",
                                    description='the command stores the content of the cell as a local file.')
        parser.add_argument(
            'file',
            type=str,
            help='filename')
        return parser

    @cell_magic
    def PYTHON(self, line, cell=None):
        """
        Defines command ``%%PYTHON``.

        .. nbref::
            :title: PYTHON

            The magic command ``%%PYTHON`` is almost to:

            ::

                with open(<filename>, "w", encoding="utf8") as f:
                    f.write("# -*- coding: utf8 -*-\\n")
                    f.write(cell.replace("\\r", ""))
        """
        parser = self.get_parser(MagicFile.PYTHON_parser, "PYTHON")
        args = self.get_args(line, parser)

        if args is not None:
            filename = args.file
            with open(filename, "w", encoding="utf8") as f:
                f.write("# -*- coding: utf8 -*-\n")
                f.write(cell.replace("\r", ""))

    @staticmethod
    def runpy_parser():
        """
        Defines the way to parse the magic command ``%%runpy``.
        """
        parser = MagicCommandParser(prog="runpy",
                                    description='run a python script which accepts standards input and ' +
                                    'produces standard outputs, a timeout is set up at 10s')
        parser.add_argument(
            'file',
            type=str,
            help='python file')
        parser.add_argument(
            'args',
            type=str,
            nargs="*",
            help='arguments for the scripts',
            default=".")
        return parser

    @cell_magic
    def runpy(self, line, cell=None):
        """
        Defines command ``%%runpy``.

        .. nbref::
            :title: runpy

            ``%%runpy`` runs  a python script which accepts
            standards input and produces standard outputs,
            a timeout is set up at 10s. It is almost equivalent to::

                from pyquickhelper.loghelper import run_cmd
                import sys
                cmd = sys.executable.replace(
                    "pythonw",
                    "python") + " " + filename + " " + args
                out, err = run_cmd(
                    cmd, wait=True, sin=cell, communicate=True, timeout=10, shell=False)

        .. versionadded:: 1.1
        """
        parser = self.get_parser(MagicFile.runpy_parser, "runpy")
        args = self.get_args(line, parser)

        if args is not None:
            filename = args.file
            if len(filename) == 0:
                self.runpy("")
            else:
                args = args.args
                cmd = sys.executable.replace("pythonw", "python")
                cmd += " " + filename + " "
                cmd += " ".join('"{0}"'.format(_)
                                for _ in args) if isinstance(args, list) else args
                tosend = cell
                out, err = run_cmd(
                    cmd, wait=True, sin=tosend, communicate=True, timeout=10, shell=False)
                if len(err) > 0:
                    return HTML(
                        '<font color="#DD0000">Error</font><br /><pre>\n%s\n</pre>' % err)
                else:
                    return HTML('<pre>\n%s\n</pre>' % out)
        return None

    @staticmethod
    def lsrepo_parser():
        """
        Defines the way to parse the magic command ``%lsrepo``.
        """
        parser = MagicCommandParser(prog="lsrepo",
                                    description='display the content of a repository (GIT or SVN)')
        parser.add_argument('path', type=str, nargs="?",
                            help='path', default=".")
        return parser

    @line_magic
    def lsrepo(self, line):
        """
        Defines ``%lsrepo``.

        .. nbref::
            :tag: file
            :title: lsrepo

            The method returns the files present in a repository
            (:epkg:`GIT` or :pkg:`SVN`). The code is equivalent to::

                from pyquickhelper.filehelper import explore_folder_iterfile_repo
                res = explore_folder_iterfile_repo(<filename>, <pattern>)
                for f in res:
                    print(f)

        .. versionadded:: 1.1
        """
        parser = self.get_parser(MagicFile.lsrepo_parser, "lsrepo")
        args = self.get_args(line, parser)

        if args is not None:
            if args.path is None or len(args.path) == 0:
                filename = "."
            else:
                filename = args.path

            iter = explore_folder_iterfile_repo(filename)
            rows = []
            for r in iter:
                d = os.path.isfile(r)
                if d:
                    st = os.stat(r)
                    r = {"name": r,
                         "size": format_file_size(st.st_size),
                         "last_modified": format_file_mtime(st.st_mtime),
                         "directory": False}
                else:
                    r = {"name": r, "directory": True}
                rows.append(r)
            return pandas.DataFrame(rows)
        return None

    @staticmethod
    def hhelp_parser():
        """
        Defines the way to parse the magic command ``%hhelp``.

        .. versionadded:: 1.1
        """
        parser = MagicCommandParser(prog="hhelp",
                                    description='display help for an object in HTML format')
        parser.add_argument(
            'obj',
            type=str,
            help='a python object')
        parser.add_argument(
            '-f',
            '--format',
            type=str,
            default="html",
            help='format',
            choices=['text', 'html', 'rst', 'rawhtml'])
        parser.add_argument(
            '-np',
            '--no-print',
            action='store_true',
            help='by default, the magic command outputs everything on the standard output, '
                 'if specified, it returns a string')
        return parser

    @line_magic
    def hhelp(self, line):
        """
        Define ``%hhelp``, it displays the help for an object in :epkg:`HTML`.

        .. nbref::
            :title: hhelp

            The magic command is equivalent to::

                from pyquickhelper import docstring2html
                docstring2html(obj, format=format)

        .. versionadded:: 1.1
        """
        parser = self.get_parser(MagicFile.hhelp_parser, "hhelp")
        args = self.get_args(line, parser)

        if args is not None:
            obj = args.obj
            format = args.format
            nop = args.no_print
            if nop or format == "html":
                return docstring2html(obj, format=format)
            else:
                print(docstring2html(obj, format=format))
        return None


def register_file_magics(ip=None):
    """
    register magics function, can be called from a notebook

    @param      ip      from ``get_ipython()``
    """
    if ip is None:
        from IPython import get_ipython
        ip = get_ipython()
    ip.register_magics(MagicFile)



# -*- coding: utf-8 -*-
"""
@file
@brief Abstract class to connect to a SQL server using various way.
It will be used to implement magic functions
"""

import re
import sqlite3


class InterfaceSQLException(BaseException):

    """
    a specific exception
    """

    def __init__(self, message):
        """
        @param      message     exception message
        """
        self.message = message

    def __str__(self):
        """
        usual
        """
        return self.message


class AutoCompletionSQLObject:

    """
    a simple class which allows auto completion
    for tables, columns...
    """

    def __init__(self, name):
        """
        creates an instance with a given name
        """
        self._true_name = name
        self._filt_name = AutoCompletionSQLObject._filter_name(name)

    @staticmethod
    def _filter_name(name):
        """
        removes unavailable characters
        """
        return name.replace(".", "_").replace(" ", "_")

    @property
    def _(self):
        """
        returns the true name of the object
        """
        return self._true_name

    @property
    def _f(self):
        """
        returns the filtered name
        """
        return self._filt_name

    def _add(self, name):
        """
        add a subname to the class

        @param      name        string
        @return                 an AutoCompletionSQLObject

        the filtered name (``_f``) of the new object will
        be added to ``self.__dict__``, if an object
        already exists with the same name, it will raise an exception
        """
        au = AutoCompletionSQLObject(name)
        af = au._f
        if af in self.__dict__:
            raise KeyError(
                "the object %s was already added to %s" %
                (af, self._f))
        self.__dict__[af] = au
        return au


class InterfaceSQL:

    """
    Abstract class to connect to a SQL server using various way.
    It will be used to implement magic functions
    """

    @staticmethod
    def create(obj):
        """

        @param      obj     a filename, a connection string, ...

        ``obj`` can be a:

        * file -->  the class :class:`Database <pyensae.sql.database_main.Database>` will be used, we assume this file
                    is sqlite database, the file does not have to exist, in that case, it will created
        * sqlite3.Connection --> the object will be wrapped into a :class:`Database <pyensae.sql.database_main.Database>`
        * InterfaceSQL --> returns the object itself

        .. versionchanged:: 1.1
            Parameter *dbfile* can be of type `sqlite3.Connection <https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection>`_.
        """
        if isinstance(obj, str):
            from .sql_interface_database import InterfaceSQLDatabase
            return InterfaceSQLDatabase(obj)
        elif isinstance(obj, sqlite3.Connection):
            from .sql_interface_database import InterfaceSQLDatabase
            return InterfaceSQLDatabase(obj)
        elif isinstance(obj, InterfaceSQL):
            return obj
        else:
            raise NotImplementedError(
                "nothing is implemented for type: %s" % str(
                    type(obj)))

    def populate_completion(self):
        """
        the method create an object which contains a class
        the user could use to speed the typing SQL queries,
        functions in a notebook

        This object will added with the name ``CC``,
        it is returned by the function.

        @return         @see cl AutoCompletionSQLObject

        The method should be called when the method @see me connect
        is called.
        """
        self.CC = AutoCompletionSQLObject("TBL")
        tbls = self.get_table_list()
        for tb in tbls:
            compl = self.CC._add(tb)
            cols = self.get_table_columns(tb)
            for k, v in cols.items():
                compl._add(v[0])
        return self.CC

    def __init__(self, obj):
        """
        Initializes the object.

        @param      obj         anything, see below

        ``obj`` can be a:
            * file -->  the class :class:`Database <pyensae.sql.database_main.Database>` will be used, we assume this file
                        is sqlite database, the file does not have to exist, in that case, it will created
        """
        raise NotImplementedError()

    def connect(self):
        """
        connection to the database
        """
        raise NotImplementedError()

    def close(self):
        """
        close the connection to the database
        """
        raise NotImplementedError()

    def get_table_list(self):
        """
        returns the list of tables in the database

        @return     list of strings
        """
        raise NotImplementedError()

    def get_table_columns(self, table_name, as_dict=True):
        """
        returns the list of columns in a table

        @param      table_name      table name
        @param      as_dict         True, as dictionary, as a list otherwise
        @return                     dictionary { "column": (position, type) } or a list
        """
        raise NotImplementedError()

    def execute(self, sql_query):
        """
        execute a SQL query

        @param      sql_query       query to execute
        @return                     pandas DataFrame

        The function takes care of the unexpected syntax introduction
        by the autocompletion object: it just replaces
        ``DB.CC.<name>`` by the ``true_name``.
        """
        sql_query = self.process_query(sql_query)
        return self.execute_clean_query(sql_query)

    def execute_clean_query(self, sql_query):
        """
        The function does the same thing as @see me execute
        but it does not replace autocompletion object.
        It is this function which should be overloaded by
        classes inheriting from this one.

        @param      sql_query       query to execute
        @return                     pandas DataFrame
        """
        raise NotImplementedError()

    _exp = re.compile("(DB[.]CC[.][a-zA-Z0-9_]+([.][a-zA-Z0-9_]+)*)")

    def process_query(self, sql_query):
        """
        replaces autocompletion object by their real names

        @param      sql_query       SQL query
        @return                     clean sql_query
        """
        # i don't remember the syntax but it should be replaced using regular expression, not
        # string replace
        fi = InterfaceSQL._exp.findall(sql_query)
        if len(fi) > 0:
            only = [_[0] for _ in fi]
            only.sort(reverse=True)
            for o in only:
                co = "self." + o[3:]
                ev = eval(co)
                sql_query = sql_query.replace(o, ev._)
        return sql_query

    def import_flat_file(self, filename, table_name):
        """
        import a flat file as a table, we assume the columns
        separator is ``\\t`` and the file name contains a header

        @param      filename        filename
        @param      table           table name
        @return                     the number of added rows
        """
        raise NotImplementedError()

    def drop_table(self, table_name):
        """
        drops a table

        @param      table           table name
        """
        raise NotImplementedError()

    def refresh_completion(self):
        """
        refresh the auto completion

        @return         completion object
        """
        return self.populate_completion()

    def add_function(self, code_function):
        """
        add a function to the database which can be called in a SELECT statement

        @param      code_function  pointer to the function
        """
        raise NotImplementedError()

    def import_dataframe(self, tablename, df):
        """
        import a dataframe into the database

        @param      tablename       name of the table
        @param      df              dataframe
        @return                     the number of added rows
        """
        raise NotImplementedError()
